from __future__ import absolute_import, division, print_function

import contextlib
import itertools
import os.path
import pickle
import shutil
import sys
import tempfile
import unittest
import warnings
from io import BytesIO

import numpy as np
import pandas as pd
import pytest

import xarray as xr
from xarray import (
    DataArray, Dataset, backends, open_dataarray, open_dataset, open_mfdataset,
    save_mfdataset)
from xarray.backends.common import (robust_getitem,
                                    PickleByReconstructionWrapper)
from xarray.backends.netCDF4_ import _extract_nc4_variable_encoding
from xarray.backends.pydap_ import PydapDataStore
from xarray.core import indexing
from xarray.core.pycompat import (
    PY2, ExitStack, basestring, dask_array_type, iteritems)
from xarray.tests import mock

from . import (
    TestCase, assert_allclose, assert_array_equal, assert_equal,
    assert_identical, has_dask, has_netCDF4, has_scipy, network, raises_regex,
    requires_dask, requires_h5netcdf, requires_netCDF4, requires_pathlib,
    requires_pydap, requires_pynio, requires_rasterio, requires_scipy,
    requires_scipy_or_netCDF4, requires_zarr, requires_pseudonetcdf,
    requires_cftime)
from .test_dataset import create_test_data

try:
    import netCDF4 as nc4
except ImportError:
    pass

try:
    import dask.array as da
except ImportError:
    pass

try:
    from pathlib import Path
except ImportError:
    try:
        from pathlib2 import Path
    except ImportError:
        pass


ON_WINDOWS = sys.platform == 'win32'


def open_example_dataset(name, *args, **kwargs):
    return open_dataset(os.path.join(os.path.dirname(__file__), 'data', name),
                        *args, **kwargs)


def open_example_mfdataset(names, *args, **kwargs):
    return open_mfdataset(
        [os.path.join(os.path.dirname(__file__), 'data', name)
         for name in names],
        *args, **kwargs)


def create_masked_and_scaled_data():
    x = np.array([np.nan, np.nan, 10, 10.1, 10.2], dtype=np.float32)
    encoding = {'_FillValue': -1, 'add_offset': 10,
                'scale_factor': np.float32(0.1), 'dtype': 'i2'}
    return Dataset({'x': ('t', x, {}, encoding)})


def create_encoded_masked_and_scaled_data():
    attributes = {'_FillValue': -1, 'add_offset': 10,
                  'scale_factor': np.float32(0.1)}
    return Dataset({'x': ('t', [-1, -1, 0, 1, 2], attributes)})


def create_unsigned_masked_scaled_data():
    encoding = {'_FillValue': 255, '_Unsigned': 'true', 'dtype': 'i1',
                'add_offset': 10, 'scale_factor': np.float32(0.1)}
    x = np.array([10.0, 10.1, 22.7, 22.8, np.nan], dtype=np.float32)
    return Dataset({'x': ('t', x, {}, encoding)})


def create_encoded_unsigned_masked_scaled_data():
    # These are values as written to the file: the _FillValue will
    # be represented in the signed form.
    attributes = {'_FillValue': -1, '_Unsigned': 'true',
                  'add_offset': 10, 'scale_factor': np.float32(0.1)}
    # Create signed data corresponding to [0, 1, 127, 128, 255] unsigned
    sb = np.asarray([0, 1, 127, -128, -1], dtype='i1')
    return Dataset({'x': ('t', sb, attributes)})


def create_boolean_data():
    attributes = {'units': '-'}
    return Dataset({'x': ('t', [True, False, False, True], attributes)})


class TestCommon(TestCase):
    def test_robust_getitem(self):

        class UnreliableArrayFailure(Exception):
            pass

        class UnreliableArray(object):
            def __init__(self, array, failures=1):
                self.array = array
                self.failures = failures

            def __getitem__(self, key):
                if self.failures > 0:
                    self.failures -= 1
                    raise UnreliableArrayFailure
                return self.array[key]

        array = UnreliableArray([0])
        with pytest.raises(UnreliableArrayFailure):
            array[0]
        self.assertEqual(array[0], 0)

        actual = robust_getitem(array, 0, catch=UnreliableArrayFailure,
                                initial_delay=0)
        self.assertEqual(actual, 0)


class NetCDF3Only(object):
    pass


class DatasetIOTestCases(object):
    autoclose = False
    engine = None
    file_format = None

    def create_store(self):
        raise NotImplementedError

    @contextlib.contextmanager
    def roundtrip(self, data, save_kwargs={}, open_kwargs={},
                  allow_cleanup_failure=False):
        with create_tmp_file(
                allow_cleanup_failure=allow_cleanup_failure) as path:
            self.save(data, path, **save_kwargs)
            with self.open(path, **open_kwargs) as ds:
                yield ds

    @contextlib.contextmanager
    def roundtrip_append(self, data, save_kwargs={}, open_kwargs={},
                         allow_cleanup_failure=False):
        with create_tmp_file(
                allow_cleanup_failure=allow_cleanup_failure) as path:
            for i, key in enumerate(data.variables):
                mode = 'a' if i > 0 else 'w'
                self.save(data[[key]], path, mode=mode, **save_kwargs)
            with self.open(path, **open_kwargs) as ds:
                yield ds

    # The save/open methods may be overwritten below
    def save(self, dataset, path, **kwargs):
        return dataset.to_netcdf(path, engine=self.engine,
                                 format=self.file_format, **kwargs)

    @contextlib.contextmanager
    def open(self, path, **kwargs):
        with open_dataset(path, engine=self.engine, autoclose=self.autoclose,
                          **kwargs) as ds:
            yield ds

    def test_zero_dimensional_variable(self):
        expected = create_test_data()
        expected['float_var'] = ([], 1.0e9, {'units': 'units of awesome'})
        expected['bytes_var'] = ([], b'foobar')
        expected['string_var'] = ([], u'foobar')
        with self.roundtrip(expected) as actual:
            assert_identical(expected, actual)

    def test_write_store(self):
        expected = create_test_data()
        with self.create_store() as store:
            expected.dump_to_store(store)
            # we need to cf decode the store because it has time and
            # non-dimension coordinates
            with xr.decode_cf(store) as actual:
                assert_allclose(expected, actual)

    def check_dtypes_roundtripped(self, expected, actual):
        for k in expected.variables:
            expected_dtype = expected.variables[k].dtype
            if (isinstance(self, NetCDF3Only) and expected_dtype == 'int64'):
                # downcast
                expected_dtype = np.dtype('int32')
            actual_dtype = actual.variables[k].dtype
            # TODO: check expected behavior for string dtypes more carefully
            string_kinds = {'O', 'S', 'U'}
            assert (expected_dtype == actual_dtype or
                    (expected_dtype.kind in string_kinds and
                     actual_dtype.kind in string_kinds))

    def test_roundtrip_test_data(self):
        expected = create_test_data()
        with self.roundtrip(expected) as actual:
            self.check_dtypes_roundtripped(expected, actual)
            assert_identical(expected, actual)

    def test_load(self):
        expected = create_test_data()

        @contextlib.contextmanager
        def assert_loads(vars=None):
            if vars is None:
                vars = expected
            with self.roundtrip(expected) as actual:
                for k, v in actual.variables.items():
                    # IndexVariables are eagerly loaded into memory
                    self.assertEqual(v._in_memory, k in actual.dims)
                yield actual
                for k, v in actual.variables.items():
                    if k in vars:
                        self.assertTrue(v._in_memory)
                assert_identical(expected, actual)

        with pytest.raises(AssertionError):
            # make sure the contextmanager works!
            with assert_loads() as ds:
                pass

        with assert_loads() as ds:
            ds.load()

        with assert_loads(['var1', 'dim1', 'dim2']) as ds:
            ds['var1'].load()

        # verify we can read data even after closing the file
        with self.roundtrip(expected) as ds:
            actual = ds.load()
        assert_identical(expected, actual)

    def test_dataset_compute(self):
        expected = create_test_data()

        with self.roundtrip(expected) as actual:
            # Test Dataset.compute()
            for k, v in actual.variables.items():
                # IndexVariables are eagerly cached
                self.assertEqual(v._in_memory, k in actual.dims)

            computed = actual.compute()

            for k, v in actual.variables.items():
                self.assertEqual(v._in_memory, k in actual.dims)
            for v in computed.variables.values():
                self.assertTrue(v._in_memory)

            assert_identical(expected, actual)
            assert_identical(expected, computed)

    def test_pickle(self):
        expected = Dataset({'foo': ('x', [42])})
        with self.roundtrip(
                expected, allow_cleanup_failure=ON_WINDOWS) as roundtripped:
            raw_pickle = pickle.dumps(roundtripped)
            # windows doesn't like opening the same file twice
            roundtripped.close()
            unpickled_ds = pickle.loads(raw_pickle)
            assert_identical(expected, unpickled_ds)

    def test_pickle_dataarray(self):
        expected = Dataset({'foo': ('x', [42])})
        with self.roundtrip(
                expected, allow_cleanup_failure=ON_WINDOWS) as roundtripped:
            unpickled_array = pickle.loads(pickle.dumps(roundtripped['foo']))
            assert_identical(expected['foo'], unpickled_array)

    def test_dataset_caching(self):
        expected = Dataset({'foo': ('x', [5, 6, 7])})
        with self.roundtrip(expected) as actual:
            assert isinstance(actual.foo.variable._data,
                              indexing.MemoryCachedArray)
            assert not actual.foo.variable._in_memory
            actual.foo.values  # cache
            assert actual.foo.variable._in_memory

        with self.roundtrip(expected, open_kwargs={'cache': False}) as actual:
            assert isinstance(actual.foo.variable._data,
                              indexing.CopyOnWriteArray)
            assert not actual.foo.variable._in_memory
            actual.foo.values  # no caching
            assert not actual.foo.variable._in_memory

    def test_roundtrip_None_variable(self):
        expected = Dataset({None: (('x', 'y'), [[0, 1], [2, 3]])})
        with self.roundtrip(expected) as actual:
            assert_identical(expected, actual)

    def test_roundtrip_object_dtype(self):
        floats = np.array([0.0, 0.0, 1.0, 2.0, 3.0], dtype=object)
        floats_nans = np.array([np.nan, np.nan, 1.0, 2.0, 3.0], dtype=object)
        bytes_ = np.array([b'ab', b'cdef', b'g'], dtype=object)
        bytes_nans = np.array([b'ab', b'cdef', np.nan], dtype=object)
        strings = np.array([u'ab', u'cdef', u'g'], dtype=object)
        strings_nans = np.array([u'ab', u'cdef', np.nan], dtype=object)
        all_nans = np.array([np.nan, np.nan], dtype=object)
        original = Dataset({'floats': ('a', floats),
                            'floats_nans': ('a', floats_nans),
                            'bytes': ('b', bytes_),
                            'bytes_nans': ('b', bytes_nans),
                            'strings': ('b', strings),
                            'strings_nans': ('b', strings_nans),
                            'all_nans': ('c', all_nans),
                            'nan': ([], np.nan)})
        expected = original.copy(deep=True)
        with self.roundtrip(original) as actual:
            try:
                assert_identical(expected, actual)
            except AssertionError:
                # Most stores use '' for nans in strings, but some don't.
                # First try the ideal case (where the store returns exactly)
                # the original Dataset), then try a more realistic case.
                # This currently includes all netCDF files when encoding is not
                # explicitly set.
                # https://github.com/pydata/xarray/issues/1647
                expected['bytes_nans'][-1] = b''
                expected['strings_nans'][-1] = u''
                assert_identical(expected, actual)

    def test_roundtrip_string_data(self):
        expected = Dataset({'x': ('t', ['ab', 'cdef'])})
        with self.roundtrip(expected) as actual:
            assert_identical(expected, actual)

    def test_roundtrip_string_encoded_characters(self):
        expected = Dataset({'x': ('t', [u'ab', u'cdef'])})
        expected['x'].encoding['dtype'] = 'S1'
        with self.roundtrip(expected) as actual:
            assert_identical(expected, actual)
            self.assertEqual(actual['x'].encoding['_Encoding'], 'utf-8')

        expected['x'].encoding['_Encoding'] = 'ascii'
        with self.roundtrip(expected) as actual:
            assert_identical(expected, actual)
            self.assertEqual(actual['x'].encoding['_Encoding'], 'ascii')

    def test_roundtrip_numpy_datetime_data(self):
        times = pd.to_datetime(['2000-01-01', '2000-01-02', 'NaT'])
        expected = Dataset({'t': ('t', times), 't0': times[0]})
        kwds = {'encoding': {'t0': {'units': 'days since 1950-01-01'}}}
        with self.roundtrip(expected, save_kwargs=kwds) as actual:
            assert_identical(expected, actual)
            assert actual.t0.encoding['units'] == 'days since 1950-01-01'

    @requires_cftime
    def test_roundtrip_cftime_datetime_data_enable_cftimeindex(self):
        from .test_coding_times import _all_cftime_date_types

        date_types = _all_cftime_date_types()
        for date_type in date_types.values():
            times = [date_type(1, 1, 1), date_type(1, 1, 2)]
            expected = Dataset({'t': ('t', times), 't0': times[0]})
            kwds = {'encoding': {'t0': {'units': 'days since 0001-01-01'}}}
            expected_decoded_t = np.array(times)
            expected_decoded_t0 = np.array([date_type(1, 1, 1)])
            expected_calendar = times[0].calendar

            with warnings.catch_warnings():
                if expected_calendar in {'proleptic_gregorian', 'gregorian'}:
                    warnings.filterwarnings(
                        'ignore', 'Unable to decode time axis')

                with xr.set_options(enable_cftimeindex=True):
                    with self.roundtrip(expected, save_kwargs=kwds) as actual:
                        abs_diff = abs(actual.t.values - expected_decoded_t)
                        assert (abs_diff <= np.timedelta64(1, 's')).all()
                        assert (actual.t.encoding['units'] ==
                                'days since 0001-01-01 00:00:00.000000')
                        assert (actual.t.encoding['calendar'] ==
                                expected_calendar)

                        abs_diff = abs(actual.t0.values - expected_decoded_t0)
                        assert (abs_diff <= np.timedelta64(1, 's')).all()
                        assert (actual.t0.encoding['units'] ==
                                'days since 0001-01-01')
                        assert (actual.t.encoding['calendar'] ==
                                expected_calendar)

    def test_roundtrip_timedelta_data(self):
        time_deltas = pd.to_timedelta(['1h', '2h', 'NaT'])
        expected = Dataset({'td': ('td', time_deltas), 'td0': time_deltas[0]})
        with self.roundtrip(expected) as actual:
            assert_identical(expected, actual)

    def test_roundtrip_float64_data(self):
        expected = Dataset({'x': ('y', np.array([1.0, 2.0, np.pi],
                                                dtype='float64'))})
        with self.roundtrip(expected) as actual:
            assert_identical(expected, actual)

    def test_roundtrip_example_1_netcdf(self):
        expected = open_example_dataset('example_1.nc')
        with self.roundtrip(expected) as actual:
            # we allow the attributes to differ since that
            # will depend on the encoding used.  For example,
            # without CF encoding 'actual' will end up with
            # a dtype attribute.
            assert_equal(expected, actual)

    def test_roundtrip_coordinates(self):
        original = Dataset({'foo': ('x', [0, 1])},
                           {'x': [2, 3], 'y': ('a', [42]), 'z': ('x', [4, 5])})

        with self.roundtrip(original) as actual:
            assert_identical(original, actual)

    def test_roundtrip_global_coordinates(self):
        original = Dataset({'x': [2, 3], 'y': ('a', [42]), 'z': ('x', [4, 5])})
        with self.roundtrip(original) as actual:
            assert_identical(original, actual)

    def test_roundtrip_coordinates_with_space(self):
        original = Dataset(coords={'x': 0, 'y z': 1})
        expected = Dataset({'y z': 1}, {'x': 0})
        with pytest.warns(xr.SerializationWarning):
            with self.roundtrip(original) as actual:
                assert_identical(expected, actual)

    def test_roundtrip_boolean_dtype(self):
        original = create_boolean_data()
        self.assertEqual(original['x'].dtype, 'bool')
        with self.roundtrip(original) as actual:
            assert_identical(original, actual)
            self.assertEqual(actual['x'].dtype, 'bool')

    def test_orthogonal_indexing(self):
        in_memory = create_test_data()
        with self.roundtrip(in_memory) as on_disk:
            indexers = {'dim1': [1, 2, 0], 'dim2': [3, 2, 0, 3],
                        'dim3': np.arange(5)}
            expected = in_memory.isel(**indexers)
            actual = on_disk.isel(**indexers)
            # make sure the array is not yet loaded into memory
            assert not actual['var1'].variable._in_memory
            assert_identical(expected, actual)
            # do it twice, to make sure we're switched from orthogonal -> numpy
            # when we cached the values
            actual = on_disk.isel(**indexers)
            assert_identical(expected, actual)

    def test_vectorized_indexing(self):
        in_memory = create_test_data()
        with self.roundtrip(in_memory) as on_disk:
            indexers = {'dim1': DataArray([0, 2, 0], dims='a'),
                        'dim2': DataArray([0, 2, 3], dims='a')}
            expected = in_memory.isel(**indexers)
            actual = on_disk.isel(**indexers)
            # make sure the array is not yet loaded into memory
            assert not actual['var1'].variable._in_memory
            assert_identical(expected, actual.load())
            # do it twice, to make sure we're switched from
            # vectorized -> numpy when we cached the values
            actual = on_disk.isel(**indexers)
            assert_identical(expected, actual)

        def multiple_indexing(indexers):
            # make sure a sequence of lazy indexings certainly works.
            with self.roundtrip(in_memory) as on_disk:
                actual = on_disk['var3']
                expected = in_memory['var3']
                for ind in indexers:
                    actual = actual.isel(**ind)
                    expected = expected.isel(**ind)
                    # make sure the array is not yet loaded into memory
                    assert not actual.variable._in_memory
                assert_identical(expected, actual.load())

        # two-staged vectorized-indexing
        indexers = [
            {'dim1': DataArray([[0, 7], [2, 6], [3, 5]], dims=['a', 'b']),
             'dim3': DataArray([[0, 4], [1, 3], [2, 2]], dims=['a', 'b'])},
            {'a': DataArray([0, 1], dims=['c']),
             'b': DataArray([0, 1], dims=['c'])}
        ]
        multiple_indexing(indexers)

        # vectorized-slice mixed
        indexers = [
            {'dim1': DataArray([[0, 7], [2, 6], [3, 5]], dims=['a', 'b']),
             'dim3': slice(None, 10)}
        ]
        multiple_indexing(indexers)

        # vectorized-integer mixed
        indexers = [
            {'dim3': 0},
            {'dim1': DataArray([[0, 7], [2, 6], [3, 5]], dims=['a', 'b'])},
            {'a': slice(None, None, 2)}
        ]
        multiple_indexing(indexers)

        # vectorized-integer mixed
        indexers = [
            {'dim3': 0},
            {'dim1': DataArray([[0, 7], [2, 6], [3, 5]], dims=['a', 'b'])},
            {'a': 1, 'b': 0}
        ]
        multiple_indexing(indexers)

        # with negative step slice.
        indexers = [
            {'dim1': DataArray([[0, 7], [2, 6], [3, 5]], dims=['a', 'b']),
             'dim3': slice(-1, 1, -1)},
        ]
        multiple_indexing(indexers)

        # with negative step slice.
        indexers = [
            {'dim1': DataArray([[0, 7], [2, 6], [3, 5]], dims=['a', 'b']),
             'dim3': slice(-1, 1, -2)},
        ]
        multiple_indexing(indexers)

    def test_isel_dataarray(self):
        # Make sure isel works lazily. GH:issue:1688
        in_memory = create_test_data()
        with self.roundtrip(in_memory) as on_disk:
            expected = in_memory.isel(dim2=in_memory['dim2'] < 3)
            actual = on_disk.isel(dim2=on_disk['dim2'] < 3)
            assert_identical(expected, actual)

    def validate_array_type(self, ds):
        # Make sure that only NumpyIndexingAdapter stores a bare np.ndarray.
        def find_and_validate_array(obj):
            # recursively called function. obj: array or array wrapper.
            if hasattr(obj, 'array'):
                if isinstance(obj.array, indexing.ExplicitlyIndexed):
                    find_and_validate_array(obj.array)
                else:
                    if isinstance(obj.array, np.ndarray):
                        assert isinstance(obj, indexing.NumpyIndexingAdapter)
                    elif isinstance(obj.array, dask_array_type):
                        assert isinstance(obj, indexing.DaskIndexingAdapter)
                    elif isinstance(obj.array, pd.Index):
                        assert isinstance(obj, indexing.PandasIndexAdapter)
                    else:
                        raise TypeError('{} is wrapped by {}'.format(
                            type(obj.array), type(obj)))

        for k, v in ds.variables.items():
            find_and_validate_array(v._data)

    def test_array_type_after_indexing(self):
        in_memory = create_test_data()
        with self.roundtrip(in_memory) as on_disk:
            self.validate_array_type(on_disk)
            indexers = {'dim1': [1, 2, 0], 'dim2': [3, 2, 0, 3],
                        'dim3': np.arange(5)}
            expected = in_memory.isel(**indexers)
            actual = on_disk.isel(**indexers)
            assert_identical(expected, actual)
            self.validate_array_type(actual)
            # do it twice, to make sure we're switched from orthogonal -> numpy
            # when we cached the values
            actual = on_disk.isel(**indexers)
            assert_identical(expected, actual)
            self.validate_array_type(actual)

    def test_dropna(self):
        # regression test for GH:issue:1694
        a = np.random.randn(4, 3)
        a[1, 1] = np.NaN
        in_memory = xr.Dataset({'a': (('y', 'x'), a)},
                               coords={'y': np.arange(4), 'x': np.arange(3)})

        assert_identical(in_memory.dropna(dim='x'),
                         in_memory.isel(x=slice(None, None, 2)))

        with self.roundtrip(in_memory) as on_disk:
            self.validate_array_type(on_disk)
            expected = in_memory.dropna(dim='x')
            actual = on_disk.dropna(dim='x')
            assert_identical(expected, actual)

    def test_ondisk_after_print(self):
        """ Make sure print does not load file into memory """
        in_memory = create_test_data()
        with self.roundtrip(in_memory) as on_disk:
            repr(on_disk)
            assert not on_disk['var1']._in_memory


class CFEncodedDataTest(DatasetIOTestCases):

    def test_roundtrip_bytes_with_fill_value(self):
        values = np.array([b'ab', b'cdef', np.nan], dtype=object)
        encoding = {'_FillValue': b'X', 'dtype': 'S1'}
        original = Dataset({'x': ('t', values, {}, encoding)})
        expected = original.copy(deep=True)
        with self.roundtrip(original) as actual:
            assert_identical(expected, actual)

        original = Dataset({'x': ('t', values, {}, {'_FillValue': b''})})
        with self.roundtrip(original) as actual:
            assert_identical(expected, actual)

    def test_roundtrip_string_with_fill_value_nchar(self):
        values = np.array([u'ab', u'cdef', np.nan], dtype=object)
        expected = Dataset({'x': ('t', values)})

        encoding = {'dtype': 'S1', '_FillValue': b'X'}
        original = Dataset({'x': ('t', values, {}, encoding)})
        # Not supported yet.
        with pytest.raises(NotImplementedError):
            with self.roundtrip(original) as actual:
                assert_identical(expected, actual)

    def test_unsigned_roundtrip_mask_and_scale(self):
        decoded = create_unsigned_masked_scaled_data()
        encoded = create_encoded_unsigned_masked_scaled_data()
        with self.roundtrip(decoded) as actual:
            for k in decoded.variables:
                self.assertEqual(decoded.variables[k].dtype,
                                 actual.variables[k].dtype)
            assert_allclose(decoded, actual, decode_bytes=False)
        with self.roundtrip(decoded,
                            open_kwargs=dict(decode_cf=False)) as actual:
            for k in encoded.variables:
                self.assertEqual(encoded.variables[k].dtype,
                                 actual.variables[k].dtype)
            assert_allclose(encoded, actual, decode_bytes=False)
        with self.roundtrip(encoded,
                            open_kwargs=dict(decode_cf=False)) as actual:
            for k in encoded.variables:
                self.assertEqual(encoded.variables[k].dtype,
                                 actual.variables[k].dtype)
            assert_allclose(encoded, actual, decode_bytes=False)
        # make sure roundtrip encoding didn't change the
        # original dataset.
        assert_allclose(
            encoded, create_encoded_unsigned_masked_scaled_data())
        with self.roundtrip(encoded) as actual:
            for k in decoded.variables:
                self.assertEqual(decoded.variables[k].dtype,
                                 actual.variables[k].dtype)
            assert_allclose(decoded, actual, decode_bytes=False)
        with self.roundtrip(encoded,
                            open_kwargs=dict(decode_cf=False)) as actual:
            for k in encoded.variables:
                self.assertEqual(encoded.variables[k].dtype,
                                 actual.variables[k].dtype)
            assert_allclose(encoded, actual, decode_bytes=False)

    def test_roundtrip_mask_and_scale(self):
        decoded = create_masked_and_scaled_data()
        encoded = create_encoded_masked_and_scaled_data()
        with self.roundtrip(decoded) as actual:
            assert_allclose(decoded, actual, decode_bytes=False)
        with self.roundtrip(decoded,
                            open_kwargs=dict(decode_cf=False)) as actual:
            # TODO: this assumes that all roundtrips will first
            # encode.  Is that something we want to test for?
            assert_allclose(encoded, actual, decode_bytes=False)
        with self.roundtrip(encoded,
                            open_kwargs=dict(decode_cf=False)) as actual:
            assert_allclose(encoded, actual, decode_bytes=False)
        # make sure roundtrip encoding didn't change the
        # original dataset.
        assert_allclose(encoded,
                        create_encoded_masked_and_scaled_data(),
                        decode_bytes=False)
        with self.roundtrip(encoded) as actual:
            assert_allclose(decoded, actual, decode_bytes=False)
        with self.roundtrip(encoded,
                            open_kwargs=dict(decode_cf=False)) as actual:
            assert_allclose(encoded, actual, decode_bytes=False)

    def test_coordinates_encoding(self):
        def equals_latlon(obj):
            return obj == 'lat lon' or obj == 'lon lat'

        original = Dataset({'temp': ('x', [0, 1]), 'precip': ('x', [0, -1])},
                           {'lat': ('x', [2, 3]), 'lon': ('x', [4, 5])})
        with self.roundtrip(original) as actual:
            assert_identical(actual, original)
        with create_tmp_file() as tmp_file:
            original.to_netcdf(tmp_file)
            with open_dataset(tmp_file, decode_coords=False) as ds:
                self.assertTrue(equals_latlon(ds['temp'].attrs['coordinates']))
                self.assertTrue(
                    equals_latlon(ds['precip'].attrs['coordinates']))
                self.assertNotIn('coordinates', ds.attrs)
                self.assertNotIn('coordinates', ds['lat'].attrs)
                self.assertNotIn('coordinates', ds['lon'].attrs)

        modified = original.drop(['temp', 'precip'])
        with self.roundtrip(modified) as actual:
            assert_identical(actual, modified)
        with create_tmp_file() as tmp_file:
            modified.to_netcdf(tmp_file)
            with open_dataset(tmp_file, decode_coords=False) as ds:
                self.assertTrue(equals_latlon(ds.attrs['coordinates']))
                self.assertNotIn('coordinates', ds['lat'].attrs)
                self.assertNotIn('coordinates', ds['lon'].attrs)

    def test_roundtrip_endian(self):
        ds = Dataset({'x': np.arange(3, 10, dtype='>i2'),
                      'y': np.arange(3, 20, dtype='<i4'),
                      'z': np.arange(3, 30, dtype='=i8'),
                      'w': ('x', np.arange(3, 10, dtype=np.float))})

        with self.roundtrip(ds) as actual:
            # technically these datasets are slightly different,
            # one hold mixed endian data (ds) the other should be
            # all big endian (actual).  assertDatasetIdentical
            # should still pass though.
            assert_identical(ds, actual)

        if self.engine == 'netcdf4':
            ds['z'].encoding['endian'] = 'big'
            with pytest.raises(NotImplementedError):
                with self.roundtrip(ds) as actual:
                    pass

    def test_invalid_dataarray_names_raise(self):
        te = (TypeError, 'string or None')
        ve = (ValueError, 'string must be length 1 or')
        data = np.random.random((2, 2))
        da = xr.DataArray(data)
        for name, e in zip([0, (4, 5), True, ''], [te, te, te, ve]):
            ds = Dataset({name: da})
            with raises_regex(*e):
                with self.roundtrip(ds):
                    pass

    def test_encoding_kwarg(self):
        ds = Dataset({'x': ('y', np.arange(10.0))})
        kwargs = dict(encoding={'x': {'dtype': 'f4'}})
        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            self.assertEqual(actual.x.encoding['dtype'], 'f4')
        self.assertEqual(ds.x.encoding, {})

        kwargs = dict(encoding={'x': {'foo': 'bar'}})
        with raises_regex(ValueError, 'unexpected encoding'):
            with self.roundtrip(ds, save_kwargs=kwargs) as actual:
                pass

        kwargs = dict(encoding={'x': 'foo'})
        with raises_regex(ValueError, 'must be castable'):
            with self.roundtrip(ds, save_kwargs=kwargs) as actual:
                pass

        kwargs = dict(encoding={'invalid': {}})
        with pytest.raises(KeyError):
            with self.roundtrip(ds, save_kwargs=kwargs) as actual:
                pass

    def test_encoding_kwarg_dates(self):
        ds = Dataset({'t': pd.date_range('2000-01-01', periods=3)})
        units = 'days since 1900-01-01'
        kwargs = dict(encoding={'t': {'units': units}})
        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            self.assertEqual(actual.t.encoding['units'], units)
            assert_identical(actual, ds)

    def test_encoding_kwarg_fixed_width_string(self):
        # regression test for GH2149
        for strings in [
            [b'foo', b'bar', b'baz'],
            [u'foo', u'bar', u'baz'],
        ]:
            ds = Dataset({'x': strings})
            kwargs = dict(encoding={'x': {'dtype': 'S1'}})
            with self.roundtrip(ds, save_kwargs=kwargs) as actual:
                self.assertEqual(actual['x'].encoding['dtype'], 'S1')
                assert_identical(actual, ds)

    def test_default_fill_value(self):
        # Test default encoding for float:
        ds = Dataset({'x': ('y', np.arange(10.0))})
        kwargs = dict(encoding={'x': {'dtype': 'f4'}})
        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            self.assertEqual(actual.x.encoding['_FillValue'],
                             np.nan)
        self.assertEqual(ds.x.encoding, {})

        # Test default encoding for int:
        ds = Dataset({'x': ('y', np.arange(10.0))})
        kwargs = dict(encoding={'x': {'dtype': 'int16'}})
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', '.*floating point data as an integer')
            with self.roundtrip(ds, save_kwargs=kwargs) as actual:
                self.assertTrue('_FillValue' not in actual.x.encoding)
        self.assertEqual(ds.x.encoding, {})

        # Test default encoding for implicit int:
        ds = Dataset({'x': ('y', np.arange(10, dtype='int16'))})
        with self.roundtrip(ds) as actual:
            self.assertTrue('_FillValue' not in actual.x.encoding)
        self.assertEqual(ds.x.encoding, {})

    def test_explicitly_omit_fill_value(self):
        ds = Dataset({'x': ('y', [np.pi, -np.pi])})
        ds.x.encoding['_FillValue'] = None
        with self.roundtrip(ds) as actual:
            assert '_FillValue' not in actual.x.encoding

    def test_explicitly_omit_fill_value_via_encoding_kwarg(self):
        ds = Dataset({'x': ('y', [np.pi, -np.pi])})
        kwargs = dict(encoding={'x': {'_FillValue': None}})
        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            assert '_FillValue' not in actual.x.encoding
        self.assertEqual(ds.y.encoding, {})

    def test_explicitly_omit_fill_value_in_coord(self):
        ds = Dataset({'x': ('y', [np.pi, -np.pi])}, coords={'y': [0.0, 1.0]})
        ds.y.encoding['_FillValue'] = None
        with self.roundtrip(ds) as actual:
            assert '_FillValue' not in actual.y.encoding

    def test_explicitly_omit_fill_value_in_coord_via_encoding_kwarg(self):
        ds = Dataset({'x': ('y', [np.pi, -np.pi])}, coords={'y': [0.0, 1.0]})
        kwargs = dict(encoding={'y': {'_FillValue': None}})
        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            assert '_FillValue' not in actual.y.encoding
        self.assertEqual(ds.y.encoding, {})

    def test_encoding_same_dtype(self):
        ds = Dataset({'x': ('y', np.arange(10.0, dtype='f4'))})
        kwargs = dict(encoding={'x': {'dtype': 'f4'}})
        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            self.assertEqual(actual.x.encoding['dtype'], 'f4')
        self.assertEqual(ds.x.encoding, {})

    def test_append_write(self):
        # regression for GH1215
        data = create_test_data()
        with self.roundtrip_append(data) as actual:
            assert_identical(data, actual)

    def test_append_overwrite_values(self):
        # regression for GH1215
        data = create_test_data()
        with create_tmp_file(allow_cleanup_failure=False) as tmp_file:
            self.save(data, tmp_file, mode='w')
            data['var2'][:] = -999
            data['var9'] = data['var2'] * 3
            self.save(data[['var2', 'var9']], tmp_file, mode='a')
            with self.open(tmp_file) as actual:
                assert_identical(data, actual)

    def test_append_with_invalid_dim_raises(self):
        data = create_test_data()
        with create_tmp_file(allow_cleanup_failure=False) as tmp_file:
            self.save(data, tmp_file, mode='w')
            data['var9'] = data['var2'] * 3
            data = data.isel(dim1=slice(2, 6))  # modify one dimension
            with raises_regex(ValueError,
                              'Unable to update size for existing dimension'):
                self.save(data, tmp_file, mode='a')

    def test_multiindex_not_implemented(self):
        ds = (Dataset(coords={'y': ('x', [1, 2]), 'z': ('x', ['a', 'b'])})
              .set_index(x=['y', 'z']))
        with raises_regex(NotImplementedError, 'MultiIndex'):
            with self.roundtrip(ds):
                pass


_counter = itertools.count()


@contextlib.contextmanager
def create_tmp_file(suffix='.nc', allow_cleanup_failure=False):
    temp_dir = tempfile.mkdtemp()
    path = os.path.join(temp_dir, 'temp-%s%s' % (next(_counter), suffix))
    try:
        yield path
    finally:
        try:
            shutil.rmtree(temp_dir)
        except OSError:
            if not allow_cleanup_failure:
                raise


@contextlib.contextmanager
def create_tmp_files(nfiles, suffix='.nc', allow_cleanup_failure=False):
    with ExitStack() as stack:
        files = [stack.enter_context(create_tmp_file(suffix,
                                                     allow_cleanup_failure))
                 for apath in np.arange(nfiles)]
        yield files


class BaseNetCDF4Test(CFEncodedDataTest):
    """Tests for both netCDF4-python and h5netcdf."""

    engine = 'netcdf4'

    def test_open_group(self):
        # Create a netCDF file with a dataset stored within a group
        with create_tmp_file() as tmp_file:
            with nc4.Dataset(tmp_file, 'w') as rootgrp:
                foogrp = rootgrp.createGroup('foo')
                ds = foogrp
                ds.createDimension('time', size=10)
                x = np.arange(10)
                ds.createVariable('x', np.int32, dimensions=('time',))
                ds.variables['x'][:] = x

            expected = Dataset()
            expected['x'] = ('time', x)

            # check equivalent ways to specify group
            for group in 'foo', '/foo', 'foo/', '/foo/':
                with self.open(tmp_file, group=group) as actual:
                    assert_equal(actual['x'], expected['x'])

            # check that missing group raises appropriate exception
            with pytest.raises(IOError):
                open_dataset(tmp_file, group='bar')
            with raises_regex(ValueError, 'must be a string'):
                open_dataset(tmp_file, group=(1, 2, 3))

    def test_open_subgroup(self):
        # Create a netCDF file with a dataset stored within a group within a
        # group
        with create_tmp_file() as tmp_file:
            rootgrp = nc4.Dataset(tmp_file, 'w')
            foogrp = rootgrp.createGroup('foo')
            bargrp = foogrp.createGroup('bar')
            ds = bargrp
            ds.createDimension('time', size=10)
            x = np.arange(10)
            ds.createVariable('x', np.int32, dimensions=('time',))
            ds.variables['x'][:] = x
            rootgrp.close()

            expected = Dataset()
            expected['x'] = ('time', x)

            # check equivalent ways to specify group
            for group in 'foo/bar', '/foo/bar', 'foo/bar/', '/foo/bar/':
                with self.open(tmp_file, group=group) as actual:
                    assert_equal(actual['x'], expected['x'])

    def test_write_groups(self):
        data1 = create_test_data()
        data2 = data1 * 2
        with create_tmp_file() as tmp_file:
            self.save(data1, tmp_file, group='data/1')
            self.save(data2, tmp_file, group='data/2', mode='a')
            with self.open(tmp_file, group='data/1') as actual1:
                assert_identical(data1, actual1)
            with self.open(tmp_file, group='data/2') as actual2:
                assert_identical(data2, actual2)

    def test_encoding_kwarg_vlen_string(self):
        for input_strings in [
            [b'foo', b'bar', b'baz'],
            [u'foo', u'bar', u'baz'],
        ]:
            original = Dataset({'x': input_strings})
            expected = Dataset({'x': [u'foo', u'bar', u'baz']})
            kwargs = dict(encoding={'x': {'dtype': str}})
            with self.roundtrip(original, save_kwargs=kwargs) as actual:
                assert actual['x'].encoding['dtype'] is str
                assert_identical(actual, expected)

    def test_roundtrip_string_with_fill_value_vlen(self):
        values = np.array([u'ab', u'cdef', np.nan], dtype=object)
        expected = Dataset({'x': ('t', values)})

        # netCDF4-based backends don't support an explicit fillvalue
        # for variable length strings yet.
        # https://github.com/Unidata/netcdf4-python/issues/730
        # https://github.com/shoyer/h5netcdf/issues/37
        original = Dataset({'x': ('t', values, {}, {'_FillValue': u'XXX'})})
        with pytest.raises(NotImplementedError):
            with self.roundtrip(original) as actual:
                assert_identical(expected, actual)

        original = Dataset({'x': ('t', values, {}, {'_FillValue': u''})})
        with pytest.raises(NotImplementedError):
            with self.roundtrip(original) as actual:
                assert_identical(expected, actual)

    def test_roundtrip_character_array(self):
        with create_tmp_file() as tmp_file:
            values = np.array([['a', 'b', 'c'], ['d', 'e', 'f']], dtype='S')

            with nc4.Dataset(tmp_file, mode='w') as nc:
                nc.createDimension('x', 2)
                nc.createDimension('string3', 3)
                v = nc.createVariable('x', np.dtype('S1'), ('x', 'string3'))
                v[:] = values

            values = np.array(['abc', 'def'], dtype='S')
            expected = Dataset({'x': ('x', values)})
            with open_dataset(tmp_file) as actual:
                assert_identical(expected, actual)
                # regression test for #157
                with self.roundtrip(actual) as roundtripped:
                    assert_identical(expected, roundtripped)

    def test_default_to_char_arrays(self):
        data = Dataset({'x': np.array(['foo', 'zzzz'], dtype='S')})
        with self.roundtrip(data) as actual:
            assert_identical(data, actual)
            self.assertEqual(actual['x'].dtype, np.dtype('S4'))

    def test_open_encodings(self):
        # Create a netCDF file with explicit time units
        # and make sure it makes it into the encodings
        # and survives a round trip
        with create_tmp_file() as tmp_file:
            with nc4.Dataset(tmp_file, 'w') as ds:
                ds.createDimension('time', size=10)
                ds.createVariable('time', np.int32, dimensions=('time',))
                units = 'days since 1999-01-01'
                ds.variables['time'].setncattr('units', units)
                ds.variables['time'][:] = np.arange(10) + 4

            expected = Dataset()

            time = pd.date_range('1999-01-05', periods=10)
            encoding = {'units': units, 'dtype': np.dtype('int32')}
            expected['time'] = ('time', time, {}, encoding)

            with open_dataset(tmp_file) as actual:
                assert_equal(actual['time'], expected['time'])
                actual_encoding = dict((k, v) for k, v in
                                       iteritems(actual['time'].encoding)
                                       if k in expected['time'].encoding)
                self.assertDictEqual(actual_encoding,
                                     expected['time'].encoding)

    def test_dump_encodings(self):
        # regression test for #709
        ds = Dataset({'x': ('y', np.arange(10.0))})
        kwargs = dict(encoding={'x': {'zlib': True}})
        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            self.assertTrue(actual.x.encoding['zlib'])

    def test_dump_and_open_encodings(self):
        # Create a netCDF file with explicit time units
        # and make sure it makes it into the encodings
        # and survives a round trip
        with create_tmp_file() as tmp_file:
            with nc4.Dataset(tmp_file, 'w') as ds:
                ds.createDimension('time', size=10)
                ds.createVariable('time', np.int32, dimensions=('time',))
                units = 'days since 1999-01-01'
                ds.variables['time'].setncattr('units', units)
                ds.variables['time'][:] = np.arange(10) + 4

            with open_dataset(tmp_file) as xarray_dataset:
                with create_tmp_file() as tmp_file2:
                    xarray_dataset.to_netcdf(tmp_file2)
                    with nc4.Dataset(tmp_file2, 'r') as ds:
                        self.assertEqual(
                            ds.variables['time'].getncattr('units'), units)
                        assert_array_equal(
                            ds.variables['time'], np.arange(10) + 4)

    def test_compression_encoding(self):
        data = create_test_data()
        data['var2'].encoding.update({'zlib': True,
                                      'chunksizes': (5, 5),
                                      'fletcher32': True,
                                      'shuffle': True,
                                      'original_shape': data.var2.shape})
        with self.roundtrip(data) as actual:
            for k, v in iteritems(data['var2'].encoding):
                self.assertEqual(v, actual['var2'].encoding[k])

        # regression test for #156
        expected = data.isel(dim1=0)
        with self.roundtrip(expected) as actual:
            assert_equal(expected, actual)

    def test_encoding_kwarg_compression(self):
        ds = Dataset({'x': np.arange(10.0)})
        encoding = dict(dtype='f4', zlib=True, complevel=9, fletcher32=True,
                        chunksizes=(5,), shuffle=True)
        kwargs = dict(encoding=dict(x=encoding))

        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            assert_equal(actual, ds)
            self.assertEqual(actual.x.encoding['dtype'], 'f4')
            self.assertEqual(actual.x.encoding['zlib'], True)
            self.assertEqual(actual.x.encoding['complevel'], 9)
            self.assertEqual(actual.x.encoding['fletcher32'], True)
            self.assertEqual(actual.x.encoding['chunksizes'], (5,))
            self.assertEqual(actual.x.encoding['shuffle'], True)

        self.assertEqual(ds.x.encoding, {})

    def test_encoding_chunksizes_unlimited(self):
        # regression test for GH1225
        ds = Dataset({'x': [1, 2, 3], 'y': ('x', [2, 3, 4])})
        ds.variables['x'].encoding = {
            'zlib': False,
            'shuffle': False,
            'complevel': 0,
            'fletcher32': False,
            'contiguous': False,
            'chunksizes': (2 ** 20,),
            'original_shape': (3,),
        }
        with self.roundtrip(ds) as actual:
            assert_equal(ds, actual)

    def test_mask_and_scale(self):
        with create_tmp_file() as tmp_file:
            with nc4.Dataset(tmp_file, mode='w') as nc:
                nc.createDimension('t', 5)
                nc.createVariable('x', 'int16', ('t',), fill_value=-1)
                v = nc.variables['x']
                v.set_auto_maskandscale(False)
                v.add_offset = 10
                v.scale_factor = 0.1
                v[:] = np.array([-1, -1, 0, 1, 2])

            # first make sure netCDF4 reads the masked and scaled data
            # correctly
            with nc4.Dataset(tmp_file, mode='r') as nc:
                expected = np.ma.array([-1, -1, 10, 10.1, 10.2],
                                       mask=[True, True, False, False, False])
                actual = nc.variables['x'][:]
                assert_array_equal(expected, actual)

            # now check xarray
            with open_dataset(tmp_file) as ds:
                expected = create_masked_and_scaled_data()
                assert_identical(expected, ds)

    def test_0dimensional_variable(self):
        # This fix verifies our work-around to this netCDF4-python bug:
        # https://github.com/Unidata/netcdf4-python/pull/220
        with create_tmp_file() as tmp_file:
            with nc4.Dataset(tmp_file, mode='w') as nc:
                v = nc.createVariable('x', 'int16')
                v[...] = 123

            with open_dataset(tmp_file) as ds:
                expected = Dataset({'x': ((), 123)})
                assert_identical(expected, ds)

    def test_already_open_dataset(self):
        with create_tmp_file() as tmp_file:
            with nc4.Dataset(tmp_file, mode='w') as nc:
                v = nc.createVariable('x', 'int')
                v[...] = 42

            nc = nc4.Dataset(tmp_file, mode='r')
            with backends.NetCDF4DataStore(nc, autoclose=False) as store:
                with open_dataset(store) as ds:
                    expected = Dataset({'x': ((), 42)})
                    assert_identical(expected, ds)

    def test_read_variable_len_strings(self):
        with create_tmp_file() as tmp_file:
            values = np.array(['foo', 'bar', 'baz'], dtype=object)

            with nc4.Dataset(tmp_file, mode='w') as nc:
                nc.createDimension('x', 3)
                v = nc.createVariable('x', str, ('x',))
                v[:] = values

            expected = Dataset({'x': ('x', values)})
            for kwargs in [{}, {'decode_cf': True}]:
                with open_dataset(tmp_file, **kwargs) as actual:
                    assert_identical(expected, actual)


@requires_netCDF4
class NetCDF4DataTest(BaseNetCDF4Test, TestCase):
    autoclose = False

    @contextlib.contextmanager
    def create_store(self):
        with create_tmp_file() as tmp_file:
            with backends.NetCDF4DataStore.open(tmp_file, mode='w') as store:
                yield store

    def test_variable_order(self):
        # doesn't work with scipy or h5py :(
        ds = Dataset()
        ds['a'] = 1
        ds['z'] = 2
        ds['b'] = 3
        ds.coords['c'] = 4

        with self.roundtrip(ds) as actual:
            self.assertEqual(list(ds.variables), list(actual.variables))

    def test_unsorted_index_raises(self):
        # should be fixed in netcdf4 v1.2.1
        random_data = np.random.random(size=(4, 6))
        dim0 = [0, 1, 2, 3]
        dim1 = [0, 2, 1, 3, 5, 4]  # We will sort this in a later step
        da = xr.DataArray(data=random_data, dims=('dim0', 'dim1'),
                          coords={'dim0': dim0, 'dim1': dim1}, name='randovar')
        ds = da.to_dataset()

        with self.roundtrip(ds) as ondisk:
            inds = np.argsort(dim1)
            ds2 = ondisk.isel(dim1=inds)
            # Older versions of NetCDF4 raise an exception here, and if so we
            # want to ensure we improve (that is, replace) the error message
            try:
                ds2.randovar.values
            except IndexError as err:
                self.assertIn('first by calling .load', str(err))

    def test_88_character_filename_segmentation_fault(self):
        # should be fixed in netcdf4 v1.3.1
        with mock.patch('netCDF4.__version__', '1.2.4'):
            with warnings.catch_warnings():
                message = ('A segmentation fault may occur when the '
                           'file path has exactly 88 characters')
                warnings.filterwarnings('error', message)
                with pytest.raises(Warning):
                    # Need to construct 88 character filepath
                    xr.Dataset().to_netcdf('a' * (88 - len(os.getcwd()) - 1))

    def test_setncattr_string(self):
        list_of_strings = ['list', 'of', 'strings']
        one_element_list_of_strings = ['one element']
        one_string = 'one string'
        attrs = {'foo': list_of_strings,
                 'bar': one_element_list_of_strings,
                 'baz': one_string}
        ds = Dataset({'x': ('y', [1, 2, 3], attrs)},
                     attrs=attrs)

        with self.roundtrip(ds) as actual:
            for totest in [actual, actual['x']]:
                assert_array_equal(list_of_strings, totest.attrs['foo'])
                assert_array_equal(one_element_list_of_strings,
                                   totest.attrs['bar'])
                assert one_string == totest.attrs['baz']


class NetCDF4DataStoreAutocloseTrue(NetCDF4DataTest):
    autoclose = True


@requires_netCDF4
@requires_dask
class NetCDF4ViaDaskDataTest(NetCDF4DataTest):
    @contextlib.contextmanager
    def roundtrip(self, data, save_kwargs={}, open_kwargs={},
                  allow_cleanup_failure=False):
        with NetCDF4DataTest.roundtrip(
                self, data, save_kwargs, open_kwargs,
                allow_cleanup_failure) as ds:
            yield ds.chunk()

    def test_unsorted_index_raises(self):
        # Skip when using dask because dask rewrites indexers to getitem,
        # dask first pulls items by block.
        pass

    def test_dataset_caching(self):
        # caching behavior differs for dask
        pass

    def test_write_inconsistent_chunks(self):
        # Construct two variables with the same dimensions, but different
        # chunk sizes.
        x = da.zeros((100, 100), dtype='f4', chunks=(50, 100))
        x = DataArray(data=x, dims=('lat', 'lon'), name='x')
        x.encoding['chunksizes'] = (50, 100)
        x.encoding['original_shape'] = (100, 100)
        y = da.ones((100, 100), dtype='f4', chunks=(100, 50))
        y = DataArray(data=y, dims=('lat', 'lon'), name='y')
        y.encoding['chunksizes'] = (100, 50)
        y.encoding['original_shape'] = (100, 100)
        # Put them both into the same dataset
        ds = Dataset({'x': x, 'y': y})
        with self.roundtrip(ds) as actual:
            assert actual['x'].encoding['chunksizes'] == (50, 100)
            assert actual['y'].encoding['chunksizes'] == (100, 50)


class NetCDF4ViaDaskDataTestAutocloseTrue(NetCDF4ViaDaskDataTest):
    autoclose = True


@requires_zarr
class BaseZarrTest(CFEncodedDataTest):

    DIMENSION_KEY = '_ARRAY_DIMENSIONS'

    @contextlib.contextmanager
    def create_store(self):
        with self.create_zarr_target() as store_target:
            yield backends.ZarrStore.open_group(store_target, mode='w')

    def save(self, dataset, store_target, **kwargs):
        return dataset.to_zarr(store=store_target, **kwargs)

    @contextlib.contextmanager
    def open(self, store_target, **kwargs):
        with xr.open_zarr(store_target, **kwargs) as ds:
            yield ds

    @contextlib.contextmanager
    def roundtrip(self, data, save_kwargs={}, open_kwargs={},
                  allow_cleanup_failure=False):
        with self.create_zarr_target() as store_target:
            self.save(data, store_target, **save_kwargs)
            with self.open(store_target, **open_kwargs) as ds:
                yield ds

    @contextlib.contextmanager
    def roundtrip_append(self, data, save_kwargs={}, open_kwargs={},
                         allow_cleanup_failure=False):
        pytest.skip("zarr backend does not support appending")

    def test_auto_chunk(self):
        original = create_test_data().chunk()

        with self.roundtrip(
                original, open_kwargs={'auto_chunk': False}) as actual:
            for k, v in actual.variables.items():
                # only index variables should be in memory
                self.assertEqual(v._in_memory, k in actual.dims)
                # there should be no chunks
                self.assertEqual(v.chunks, None)

        with self.roundtrip(
                original, open_kwargs={'auto_chunk': True}) as actual:
            for k, v in actual.variables.items():
                # only index variables should be in memory
                self.assertEqual(v._in_memory, k in actual.dims)
                # chunk size should be the same as original
                self.assertEqual(v.chunks, original[k].chunks)

    def test_write_uneven_dask_chunks(self):
        # regression for GH#2225
        original = create_test_data().chunk({'dim1': 3, 'dim2': 4, 'dim3': 3})

        with self.roundtrip(
                original, open_kwargs={'auto_chunk': True}) as actual:
            for k, v in actual.data_vars.items():
                print(k)
                assert v.chunks == actual[k].chunks

    def test_chunk_encoding(self):
        # These datasets have no dask chunks. All chunking specified in
        # encoding
        data = create_test_data()
        chunks = (5, 5)
        data['var2'].encoding.update({'chunks': chunks})

        with self.roundtrip(data) as actual:
            self.assertEqual(chunks, actual['var2'].encoding['chunks'])

        # expect an error with non-integer chunks
        data['var2'].encoding.update({'chunks': (5, 4.5)})
        with pytest.raises(TypeError):
            with self.roundtrip(data) as actual:
                pass

    def test_chunk_encoding_with_dask(self):
        # These datasets DO have dask chunks. Need to check for various
        # interactions between dask and zarr chunks
        ds = xr.DataArray((np.arange(12)), dims='x', name='var1').to_dataset()

        # - no encoding specified -
        # zarr automatically gets chunk information from dask chunks
        ds_chunk4 = ds.chunk({'x': 4})
        with self.roundtrip(ds_chunk4) as actual:
            self.assertEqual((4,), actual['var1'].encoding['chunks'])

        # should fail if dask_chunks are irregular...
        ds_chunk_irreg = ds.chunk({'x': (5, 4, 3)})
        with pytest.raises(ValueError) as e_info:
            with self.roundtrip(ds_chunk_irreg) as actual:
                pass
        # make sure this error message is correct and not some other error
        assert e_info.match('chunks')

        # ... except if the last chunk is smaller than the first
        ds_chunk_irreg = ds.chunk({'x': (5, 5, 2)})
        with self.roundtrip(ds_chunk_irreg) as actual:
            self.assertEqual((5,), actual['var1'].encoding['chunks'])

        # - encoding specified  -
        # specify compatible encodings
        for chunk_enc in 4, (4, ):
            ds_chunk4['var1'].encoding.update({'chunks': chunk_enc})
            with self.roundtrip(ds_chunk4) as actual:
                self.assertEqual((4,), actual['var1'].encoding['chunks'])

        # specify incompatible encoding
        ds_chunk4['var1'].encoding.update({'chunks': (5, 5)})
        with pytest.raises(ValueError) as e_info:
            with self.roundtrip(ds_chunk4) as actual:
                pass
        assert e_info.match('chunks')

        # TODO: remove this failure once syncronized overlapping writes are
        # supported by xarray
        ds_chunk4['var1'].encoding.update({'chunks': 5})
        with pytest.raises(NotImplementedError):
            with self.roundtrip(ds_chunk4) as actual:
                pass

    def test_hidden_zarr_keys(self):
        expected = create_test_data()
        with self.create_store() as store:
            expected.dump_to_store(store)
            zarr_group = store.ds

            # check that a variable hidden attribute is present and correct
            # JSON only has a single array type, which maps to list in Python.
            # In contrast, dims in xarray is always a tuple.
            for var in expected.variables.keys():
                dims = zarr_group[var].attrs[self.DIMENSION_KEY]
                assert dims == list(expected[var].dims)

            with xr.decode_cf(store):
                # make sure it is hidden
                for var in expected.variables.keys():
                    assert self.DIMENSION_KEY not in expected[var].attrs

            # put it back and try removing from a variable
            del zarr_group.var2.attrs[self.DIMENSION_KEY]
            with pytest.raises(KeyError):
                with xr.decode_cf(store):
                    pass

    def test_write_persistence_modes(self):
        original = create_test_data()

        # overwrite mode
        with self.roundtrip(original, save_kwargs={'mode': 'w'}) as actual:
            assert_identical(original, actual)

        # don't overwrite mode
        with self.roundtrip(original, save_kwargs={'mode': 'w-'}) as actual:
            assert_identical(original, actual)

        # make sure overwriting works as expected
        with self.create_zarr_target() as store:
            self.save(original, store)
            # should overwrite with no error
            self.save(original, store, mode='w')
            with self.open(store) as actual:
                assert_identical(original, actual)
                with pytest.raises(ValueError):
                    self.save(original, store, mode='w-')

        # check that we can't use other persistence modes
        # TODO: reconsider whether other persistence modes should be supported
        with pytest.raises(ValueError):
            with self.roundtrip(original, save_kwargs={'mode': 'a'}) as actual:
                pass

    def test_compressor_encoding(self):
        original = create_test_data()
        # specify a custom compressor
        import zarr
        blosc_comp = zarr.Blosc(cname='zstd', clevel=3, shuffle=2)
        save_kwargs = dict(encoding={'var1': {'compressor': blosc_comp}})
        with self.roundtrip(original, save_kwargs=save_kwargs) as ds:
            actual = ds['var1'].encoding['compressor']
            # get_config returns a dictionary of compressor attributes
            assert actual.get_config() == blosc_comp.get_config()

    def test_group(self):
        original = create_test_data()
        group = 'some/random/path'
        with self.roundtrip(original, save_kwargs={'group': group},
                            open_kwargs={'group': group}) as actual:
            assert_identical(original, actual)

    def test_encoding_kwarg_fixed_width_string(self):
        # not relevant for zarr, since we don't use EncodedStringCoder
        pass

    # TODO: someone who understand caching figure out whether chaching
    # makes sense for Zarr backend
    @pytest.mark.xfail(reason="Zarr caching not implemented")
    def test_dataset_caching(self):
        super(CFEncodedDataTest, self).test_dataset_caching()

    @pytest.mark.xfail(reason="Zarr stores can not be appended to")
    def test_append_write(self):
        super(CFEncodedDataTest, self).test_append_write()

    @pytest.mark.xfail(reason="Zarr stores can not be appended to")
    def test_append_overwrite_values(self):
        super(CFEncodedDataTest, self).test_append_overwrite_values()

    @pytest.mark.xfail(reason="Zarr stores can not be appended to")
    def test_append_with_invalid_dim_raises(self):
        super(CFEncodedDataTest, self).test_append_with_invalid_dim_raises()

    def test_to_zarr_compute_false_roundtrip(self):
        from dask.delayed import Delayed

        original = create_test_data().chunk()

        with self.create_zarr_target() as store:
            delayed_obj = self.save(original, store, compute=False)
            assert isinstance(delayed_obj, Delayed)
            delayed_obj.compute()

            with self.open(store) as actual:
                assert_identical(original, actual)


@requires_zarr
class ZarrDictStoreTest(BaseZarrTest, TestCase):
    @contextlib.contextmanager
    def create_zarr_target(self):
        yield {}


@requires_zarr
class ZarrDirectoryStoreTest(BaseZarrTest, TestCase):
    @contextlib.contextmanager
    def create_zarr_target(self):
        with create_tmp_file(suffix='.zarr') as tmp:
            yield tmp


class ScipyWriteTest(CFEncodedDataTest, NetCDF3Only):

    def test_append_write(self):
        import scipy
        if scipy.__version__ == '1.0.1':
            pytest.xfail('https://github.com/scipy/scipy/issues/8625')
        super(ScipyWriteTest, self).test_append_write()

    def test_append_overwrite_values(self):
        import scipy
        if scipy.__version__ == '1.0.1':
            pytest.xfail('https://github.com/scipy/scipy/issues/8625')
        super(ScipyWriteTest, self).test_append_overwrite_values()


@requires_scipy
class ScipyInMemoryDataTest(ScipyWriteTest, TestCase):
    engine = 'scipy'

    @contextlib.contextmanager
    def create_store(self):
        fobj = BytesIO()
        yield backends.ScipyDataStore(fobj, 'w')

    def test_to_netcdf_explicit_engine(self):
        # regression test for GH1321
        Dataset({'foo': 42}).to_netcdf(engine='scipy')

    @pytest.mark.skipif(PY2, reason='cannot pickle BytesIO on Python 2')
    def test_bytesio_pickle(self):
        data = Dataset({'foo': ('x', [1, 2, 3])})
        fobj = BytesIO(data.to_netcdf())
        with open_dataset(fobj, autoclose=self.autoclose) as ds:
            unpickled = pickle.loads(pickle.dumps(ds))
            assert_identical(unpickled, data)


class ScipyInMemoryDataTestAutocloseTrue(ScipyInMemoryDataTest):
    autoclose = True


@requires_scipy
class ScipyFileObjectTest(ScipyWriteTest, TestCase):
    engine = 'scipy'

    @contextlib.contextmanager
    def create_store(self):
        fobj = BytesIO()
        yield backends.ScipyDataStore(fobj, 'w')

    @contextlib.contextmanager
    def roundtrip(self, data, save_kwargs={}, open_kwargs={},
                  allow_cleanup_failure=False):
        with create_tmp_file() as tmp_file:
            with open(tmp_file, 'wb') as f:
                self.save(data, f, **save_kwargs)
            with open(tmp_file, 'rb') as f:
                with self.open(f, **open_kwargs) as ds:
                    yield ds

    @pytest.mark.skip(reason='cannot pickle file objects')
    def test_pickle(self):
        pass

    @pytest.mark.skip(reason='cannot pickle file objects')
    def test_pickle_dataarray(self):
        pass


@requires_scipy
class ScipyFilePathTest(ScipyWriteTest, TestCase):
    engine = 'scipy'

    @contextlib.contextmanager
    def create_store(self):
        with create_tmp_file() as tmp_file:
            with backends.ScipyDataStore(tmp_file, mode='w') as store:
                yield store

    def test_array_attrs(self):
        ds = Dataset(attrs={'foo': [[1, 2], [3, 4]]})
        with raises_regex(ValueError, 'must be 1-dimensional'):
            with self.roundtrip(ds):
                pass

    def test_roundtrip_example_1_netcdf_gz(self):
        with open_example_dataset('example_1.nc.gz') as expected:
            with open_example_dataset('example_1.nc') as actual:
                assert_identical(expected, actual)

    def test_netcdf3_endianness(self):
        # regression test for GH416
        expected = open_example_dataset('bears.nc', engine='scipy')
        for var in expected.variables.values():
            self.assertTrue(var.dtype.isnative)

    @requires_netCDF4
    def test_nc4_scipy(self):
        with create_tmp_file(allow_cleanup_failure=True) as tmp_file:
            with nc4.Dataset(tmp_file, 'w', format='NETCDF4') as rootgrp:
                rootgrp.createGroup('foo')

            with raises_regex(TypeError, 'pip install netcdf4'):
                open_dataset(tmp_file, engine='scipy')


class ScipyFilePathTestAutocloseTrue(ScipyFilePathTest):
    autoclose = True


@requires_netCDF4
class NetCDF3ViaNetCDF4DataTest(CFEncodedDataTest, NetCDF3Only, TestCase):
    engine = 'netcdf4'
    file_format = 'NETCDF3_CLASSIC'

    @contextlib.contextmanager
    def create_store(self):
        with create_tmp_file() as tmp_file:
            with backends.NetCDF4DataStore.open(
                    tmp_file, mode='w', format='NETCDF3_CLASSIC') as store:
                yield store

    def test_encoding_kwarg_vlen_string(self):
        original = Dataset({'x': [u'foo', u'bar', u'baz']})
        kwargs = dict(encoding={'x': {'dtype': str}})
        with raises_regex(ValueError, 'encoding dtype=str for vlen'):
            with self.roundtrip(original, save_kwargs=kwargs):
                pass


class NetCDF3ViaNetCDF4DataTestAutocloseTrue(NetCDF3ViaNetCDF4DataTest):
    autoclose = True


@requires_netCDF4
class NetCDF4ClassicViaNetCDF4DataTest(CFEncodedDataTest, NetCDF3Only,
                                       TestCase):
    engine = 'netcdf4'
    file_format = 'NETCDF4_CLASSIC'

    @contextlib.contextmanager
    def create_store(self):
        with create_tmp_file() as tmp_file:
            with backends.NetCDF4DataStore.open(
                    tmp_file, mode='w', format='NETCDF4_CLASSIC') as store:
                yield store


class NetCDF4ClassicViaNetCDF4DataTestAutocloseTrue(
        NetCDF4ClassicViaNetCDF4DataTest):
    autoclose = True


@requires_scipy_or_netCDF4
class GenericNetCDFDataTest(CFEncodedDataTest, NetCDF3Only, TestCase):
    # verify that we can read and write netCDF3 files as long as we have scipy
    # or netCDF4-python installed
    file_format = 'netcdf3_64bit'

    def test_write_store(self):
        # there's no specific store to test here
        pass

    def test_engine(self):
        data = create_test_data()
        with raises_regex(ValueError, 'unrecognized engine'):
            data.to_netcdf('foo.nc', engine='foobar')
        with raises_regex(ValueError, 'invalid engine'):
            data.to_netcdf(engine='netcdf4')

        with create_tmp_file() as tmp_file:
            data.to_netcdf(tmp_file)
            with raises_regex(ValueError, 'unrecognized engine'):
                open_dataset(tmp_file, engine='foobar')

        netcdf_bytes = data.to_netcdf()
        with raises_regex(ValueError, 'can only read'):
            open_dataset(BytesIO(netcdf_bytes), engine='foobar')

    @pytest.mark.xfail(reason='https://github.com/pydata/xarray/issues/2050')
    def test_cross_engine_read_write_netcdf3(self):
        data = create_test_data()
        valid_engines = set()
        if has_netCDF4:
            valid_engines.add('netcdf4')
        if has_scipy:
            valid_engines.add('scipy')

        for write_engine in valid_engines:
            for format in ['NETCDF3_CLASSIC', 'NETCDF3_64BIT']:
                with create_tmp_file() as tmp_file:
                    data.to_netcdf(tmp_file, format=format,
                                   engine=write_engine)
                    for read_engine in valid_engines:
                        with open_dataset(tmp_file,
                                          engine=read_engine) as actual:
                            # hack to allow test to work:
                            # coord comes back as DataArray rather than coord,
                            # and so need to loop through here rather than in
                            # the test function (or we get recursion)
                            [assert_allclose(data[k].variable,
                                             actual[k].variable)
                             for k in data.variables]

    def test_encoding_unlimited_dims(self):
        ds = Dataset({'x': ('y', np.arange(10.0))})
        with self.roundtrip(ds,
                            save_kwargs=dict(unlimited_dims=['y'])) as actual:
            self.assertEqual(actual.encoding['unlimited_dims'], set('y'))
            assert_equal(ds, actual)

        # Regression test for https://github.com/pydata/xarray/issues/2134
        with self.roundtrip(ds,
                            save_kwargs=dict(unlimited_dims='y')) as actual:
            self.assertEqual(actual.encoding['unlimited_dims'], set('y'))
            assert_equal(ds, actual)

        ds.encoding = {'unlimited_dims': ['y']}
        with self.roundtrip(ds) as actual:
            self.assertEqual(actual.encoding['unlimited_dims'], set('y'))
            assert_equal(ds, actual)

        # Regression test for https://github.com/pydata/xarray/issues/2134
        ds.encoding = {'unlimited_dims': 'y'}
        with self.roundtrip(ds) as actual:
            self.assertEqual(actual.encoding['unlimited_dims'], set('y'))
            assert_equal(ds, actual)


class GenericNetCDFDataTestAutocloseTrue(GenericNetCDFDataTest):
    autoclose = True


@requires_h5netcdf
@requires_netCDF4
class H5NetCDFDataTest(BaseNetCDF4Test, TestCase):
    engine = 'h5netcdf'

    @contextlib.contextmanager
    def create_store(self):
        with create_tmp_file() as tmp_file:
            yield backends.H5NetCDFStore(tmp_file, 'w')

    def test_complex(self):
        expected = Dataset({'x': ('y', np.ones(5) + 1j * np.ones(5))})
        with self.roundtrip(expected) as actual:
            assert_equal(expected, actual)

    @pytest.mark.xfail(reason='https://github.com/pydata/xarray/issues/535')
    def test_cross_engine_read_write_netcdf4(self):
        # Drop dim3, because its labels include strings. These appear to be
        # not properly read with python-netCDF4, which converts them into
        # unicode instead of leaving them as bytes.
        data = create_test_data().drop('dim3')
        data.attrs['foo'] = 'bar'
        valid_engines = ['netcdf4', 'h5netcdf']
        for write_engine in valid_engines:
            with create_tmp_file() as tmp_file:
                data.to_netcdf(tmp_file, engine=write_engine)
                for read_engine in valid_engines:
                    with open_dataset(tmp_file, engine=read_engine) as actual:
                        assert_identical(data, actual)

    def test_read_byte_attrs_as_unicode(self):
        with create_tmp_file() as tmp_file:
            with nc4.Dataset(tmp_file, 'w') as nc:
                nc.foo = b'bar'
            with open_dataset(tmp_file) as actual:
                expected = Dataset(attrs={'foo': 'bar'})
                assert_identical(expected, actual)

    def test_encoding_unlimited_dims(self):
        ds = Dataset({'x': ('y', np.arange(10.0))})
        with self.roundtrip(ds,
                            save_kwargs=dict(unlimited_dims=['y'])) as actual:
            self.assertEqual(actual.encoding['unlimited_dims'], set('y'))
            assert_equal(ds, actual)
        ds.encoding = {'unlimited_dims': ['y']}
        with self.roundtrip(ds) as actual:
            self.assertEqual(actual.encoding['unlimited_dims'], set('y'))
            assert_equal(ds, actual)

    def test_compression_encoding_h5py(self):
        ENCODINGS = (
            # h5py style compression with gzip codec will be converted to
            # NetCDF4-Python style on round-trip
            ({'compression': 'gzip', 'compression_opts': 9},
             {'zlib': True, 'complevel': 9}),
            # What can't be expressed in NetCDF4-Python style is
            # round-tripped unaltered
            ({'compression': 'lzf', 'compression_opts': None},
             {'compression': 'lzf', 'compression_opts': None}),
            # If both styles are used together, h5py format takes precedence
            ({'compression': 'lzf', 'compression_opts': None,
              'zlib': True, 'complevel': 9},
             {'compression': 'lzf', 'compression_opts': None}))

        for compr_in, compr_out in ENCODINGS:
            data = create_test_data()
            compr_common = {
                'chunksizes': (5, 5),
                'fletcher32': True,
                'shuffle': True,
                'original_shape': data.var2.shape
            }
            data['var2'].encoding.update(compr_in)
            data['var2'].encoding.update(compr_common)
            compr_out.update(compr_common)
            with self.roundtrip(data) as actual:
                for k, v in compr_out.items():
                    self.assertEqual(v, actual['var2'].encoding[k])

    def test_compression_check_encoding_h5py(self):
        """When mismatched h5py and NetCDF4-Python encodings are expressed
        in to_netcdf(encoding=...), must raise ValueError
        """
        data = Dataset({'x': ('y', np.arange(10.0))})
        # Compatible encodings are graciously supported
        with create_tmp_file() as tmp_file:
            data.to_netcdf(
                tmp_file, engine='h5netcdf',
                encoding={'x': {'compression': 'gzip', 'zlib': True,
                                'compression_opts': 6, 'complevel': 6}})
            with open_dataset(tmp_file, engine='h5netcdf') as actual:
                assert actual.x.encoding['zlib'] is True
                assert actual.x.encoding['complevel'] == 6

        # Incompatible encodings cause a crash
        with create_tmp_file() as tmp_file:
            with raises_regex(ValueError,
                              "'zlib' and 'compression' encodings mismatch"):
                data.to_netcdf(
                    tmp_file, engine='h5netcdf',
                    encoding={'x': {'compression': 'lzf', 'zlib': True}})

        with create_tmp_file() as tmp_file:
            with raises_regex(
                    ValueError,
                    "'complevel' and 'compression_opts' encodings mismatch"):
                data.to_netcdf(
                    tmp_file, engine='h5netcdf',
                    encoding={'x': {'compression': 'gzip',
                                    'compression_opts': 5, 'complevel': 6}})

    def test_dump_encodings_h5py(self):
        # regression test for #709
        ds = Dataset({'x': ('y', np.arange(10.0))})

        kwargs = {'encoding': {'x': {
            'compression': 'gzip', 'compression_opts': 9}}}
        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            self.assertEqual(actual.x.encoding['zlib'], True)
            self.assertEqual(actual.x.encoding['complevel'], 9)

        kwargs = {'encoding': {'x': {
            'compression': 'lzf', 'compression_opts': None}}}
        with self.roundtrip(ds, save_kwargs=kwargs) as actual:
            self.assertEqual(actual.x.encoding['compression'], 'lzf')
            self.assertEqual(actual.x.encoding['compression_opts'], None)


# tests pending h5netcdf fix
@unittest.skip
class H5NetCDFDataTestAutocloseTrue(H5NetCDFDataTest):
    autoclose = True


@pytest.fixture(params=['scipy', 'netcdf4', 'h5netcdf', 'pynio'])
def readengine(request):
    return request.param


@pytest.fixture(params=[1, 100])
def nfiles(request):
    return request.param


@pytest.fixture(params=[True, False])
def autoclose(request):
    return request.param


@pytest.fixture(params=[True, False])
def parallel(request):
    return request.param


@pytest.fixture(params=[None, 5])
def chunks(request):
    return request.param


# using pytest.mark.skipif does not work so this a work around
def skip_if_not_engine(engine):
    if engine == 'netcdf4':
        pytest.importorskip('netCDF4')
    elif engine == 'pynio':
        pytest.importorskip('Nio')
    else:
        pytest.importorskip(engine)


def test_open_mfdataset_manyfiles(readengine, nfiles, autoclose, parallel,
                                  chunks):

    # skip certain combinations
    skip_if_not_engine(readengine)

    if not has_dask and parallel:
        pytest.skip('parallel requires dask')

    if readengine == 'h5netcdf' and autoclose:
        pytest.skip('h5netcdf does not support autoclose yet')

    if ON_WINDOWS:
        pytest.skip('Skipping on Windows')

    randdata = np.random.randn(nfiles)
    original = Dataset({'foo': ('x', randdata)})
    # test standard open_mfdataset approach with too many files
    with create_tmp_files(nfiles) as tmpfiles:
        writeengine = (readengine if readengine != 'pynio' else 'netcdf4')
        # split into multiple sets of temp files
        for ii in original.x.values:
            subds = original.isel(x=slice(ii, ii + 1))
            subds.to_netcdf(tmpfiles[ii], engine=writeengine)

        # check that calculation on opened datasets works properly
        actual = open_mfdataset(tmpfiles, engine=readengine, parallel=parallel,
                                autoclose=autoclose, chunks=chunks)

        # check that using open_mfdataset returns dask arrays for variables
        assert isinstance(actual['foo'].data, dask_array_type)

        assert_identical(original, actual)


@requires_scipy_or_netCDF4
class OpenMFDatasetWithDataVarsAndCoordsKwTest(TestCase):
    coord_name = 'lon'
    var_name = 'v1'

    @contextlib.contextmanager
    def setup_files_and_datasets(self):
        ds1, ds2 = self.gen_datasets_with_common_coord_and_time()
        with create_tmp_file() as tmpfile1:
            with create_tmp_file() as tmpfile2:

                # save data to the temporary files
                ds1.to_netcdf(tmpfile1)
                ds2.to_netcdf(tmpfile2)

                yield [tmpfile1, tmpfile2], [ds1, ds2]

    def gen_datasets_with_common_coord_and_time(self):
        # create coordinate data
        nx = 10
        nt = 10
        x = np.arange(nx)
        t1 = np.arange(nt)
        t2 = np.arange(nt, 2 * nt, 1)

        v1 = np.random.randn(nt, nx)
        v2 = np.random.randn(nt, nx)

        ds1 = Dataset(data_vars={self.var_name: (['t', 'x'], v1),
                                 self.coord_name: ('x', 2 * x)},
                      coords={
                          't': (['t', ], t1),
                          'x': (['x', ], x)
        })

        ds2 = Dataset(data_vars={self.var_name: (['t', 'x'], v2),
                                 self.coord_name: ('x', 2 * x)},
                      coords={
                          't': (['t', ], t2),
                          'x': (['x', ], x)
        })

        return ds1, ds2

    def test_open_mfdataset_does_same_as_concat(self):
        options = ['all', 'minimal', 'different', ]

        with self.setup_files_and_datasets() as (files, [ds1, ds2]):
            for opt in options:
                with open_mfdataset(files, data_vars=opt) as ds:
                    kwargs = dict(data_vars=opt, dim='t')
                    ds_expect = xr.concat([ds1, ds2], **kwargs)
                    assert_identical(ds, ds_expect)

                with open_mfdataset(files, coords=opt) as ds:
                    kwargs = dict(coords=opt, dim='t')
                    ds_expect = xr.concat([ds1, ds2], **kwargs)
                    assert_identical(ds, ds_expect)

    def test_common_coord_when_datavars_all(self):
        opt = 'all'

        with self.setup_files_and_datasets() as (files, [ds1, ds2]):
            # open the files with the data_var option
            with open_mfdataset(files, data_vars=opt) as ds:

                coord_shape = ds[self.coord_name].shape
                coord_shape1 = ds1[self.coord_name].shape
                coord_shape2 = ds2[self.coord_name].shape

                var_shape = ds[self.var_name].shape

                self.assertEqual(var_shape, coord_shape)
                self.assertNotEqual(coord_shape1, coord_shape)
                self.assertNotEqual(coord_shape2, coord_shape)

    def test_common_coord_when_datavars_minimal(self):
        opt = 'minimal'

        with self.setup_files_and_datasets() as (files, [ds1, ds2]):
            # open the files using data_vars option
            with open_mfdataset(files, data_vars=opt) as ds:

                coord_shape = ds[self.coord_name].shape
                coord_shape1 = ds1[self.coord_name].shape
                coord_shape2 = ds2[self.coord_name].shape

                var_shape = ds[self.var_name].shape

                self.assertNotEqual(var_shape, coord_shape)
                self.assertEqual(coord_shape1, coord_shape)
                self.assertEqual(coord_shape2, coord_shape)

    def test_invalid_data_vars_value_should_fail(self):

        with self.setup_files_and_datasets() as (files, _):
            with pytest.raises(ValueError):
                with open_mfdataset(files, data_vars='minimum'):
                    pass

            # test invalid coord parameter
            with pytest.raises(ValueError):
                with open_mfdataset(files, coords='minimum'):
                    pass


@requires_dask
@requires_scipy
@requires_netCDF4
class DaskTest(TestCase, DatasetIOTestCases):
    @contextlib.contextmanager
    def create_store(self):
        yield Dataset()

    @contextlib.contextmanager
    def roundtrip(self, data, save_kwargs={}, open_kwargs={},
                  allow_cleanup_failure=False):
        yield data.chunk()

    # Override methods in DatasetIOTestCases - not applicable to dask
    def test_roundtrip_string_encoded_characters(self):
        pass

    def test_roundtrip_coordinates_with_space(self):
        pass

    def test_roundtrip_numpy_datetime_data(self):
        # Override method in DatasetIOTestCases - remove not applicable
        # save_kwds
        times = pd.to_datetime(['2000-01-01', '2000-01-02', 'NaT'])
        expected = Dataset({'t': ('t', times), 't0': times[0]})
        with self.roundtrip(expected) as actual:
            assert_identical(expected, actual)

    def test_roundtrip_cftime_datetime_data_enable_cftimeindex(self):
        # Override method in DatasetIOTestCases - remove not applicable
        # save_kwds
        from .test_coding_times import _all_cftime_date_types

        date_types = _all_cftime_date_types()
        for date_type in date_types.values():
            times = [date_type(1, 1, 1), date_type(1, 1, 2)]
            expected = Dataset({'t': ('t', times), 't0': times[0]})
            expected_decoded_t = np.array(times)
            expected_decoded_t0 = np.array([date_type(1, 1, 1)])

            with xr.set_options(enable_cftimeindex=True):
                with self.roundtrip(expected) as actual:
                    abs_diff = abs(actual.t.values - expected_decoded_t)
                    self.assertTrue((abs_diff <= np.timedelta64(1, 's')).all())

                    abs_diff = abs(actual.t0.values - expected_decoded_t0)
                    self.assertTrue((abs_diff <= np.timedelta64(1, 's')).all())

    def test_roundtrip_cftime_datetime_data_disable_cftimeindex(self):
        # Override method in DatasetIOTestCases - remove not applicable
        # save_kwds
        from .test_coding_times import _all_cftime_date_types

        date_types = _all_cftime_date_types()
        for date_type in date_types.values():
            times = [date_type(1, 1, 1), date_type(1, 1, 2)]
            expected = Dataset({'t': ('t', times), 't0': times[0]})
            expected_decoded_t = np.array(times)
            expected_decoded_t0 = np.array([date_type(1, 1, 1)])

            with xr.set_options(enable_cftimeindex=False):
                with self.roundtrip(expected) as actual:
                    abs_diff = abs(actual.t.values - expected_decoded_t)
                    self.assertTrue((abs_diff <= np.timedelta64(1, 's')).all())

                    abs_diff = abs(actual.t0.values - expected_decoded_t0)
                    self.assertTrue((abs_diff <= np.timedelta64(1, 's')).all())

    def test_write_store(self):
        # Override method in DatasetIOTestCases - not applicable to dask
        pass

    def test_dataset_caching(self):
        expected = Dataset({'foo': ('x', [5, 6, 7])})
        with self.roundtrip(expected) as actual:
            assert not actual.foo.variable._in_memory
            actual.foo.values  # no caching
            assert not actual.foo.variable._in_memory

    def test_open_mfdataset(self):
        original = Dataset({'foo': ('x', np.random.randn(10))})
        with create_tmp_file() as tmp1:
            with create_tmp_file() as tmp2:
                original.isel(x=slice(5)).to_netcdf(tmp1)
                original.isel(x=slice(5, 10)).to_netcdf(tmp2)
                with open_mfdataset([tmp1, tmp2],
                                    autoclose=self.autoclose) as actual:
                    self.assertIsInstance(actual.foo.variable.data, da.Array)
                    self.assertEqual(actual.foo.variable.data.chunks,
                                     ((5, 5),))
                    assert_identical(original, actual)
                with open_mfdataset([tmp1, tmp2], chunks={'x': 3},
                                    autoclose=self.autoclose) as actual:
                    self.assertEqual(actual.foo.variable.data.chunks,
                                     ((3, 2, 3, 2),))

        with raises_regex(IOError, 'no files to open'):
            open_mfdataset('foo-bar-baz-*.nc', autoclose=self.autoclose)

        with raises_regex(ValueError, 'wild-card'):
            open_mfdataset('http://some/remote/uri', autoclose=self.autoclose)

    @requires_pathlib
    def test_open_mfdataset_pathlib(self):
        original = Dataset({'foo': ('x', np.random.randn(10))})
        with create_tmp_file() as tmp1:
            with create_tmp_file() as tmp2:
                tmp1 = Path(tmp1)
                tmp2 = Path(tmp2)
                original.isel(x=slice(5)).to_netcdf(tmp1)
                original.isel(x=slice(5, 10)).to_netcdf(tmp2)
                with open_mfdataset([tmp1, tmp2],
                                    autoclose=self.autoclose) as actual:
                    assert_identical(original, actual)

    def test_attrs_mfdataset(self):
        original = Dataset({'foo': ('x', np.random.randn(10))})
        with create_tmp_file() as tmp1:
            with create_tmp_file() as tmp2:
                ds1 = original.isel(x=slice(5))
                ds2 = original.isel(x=slice(5, 10))
                ds1.attrs['test1'] = 'foo'
                ds2.attrs['test2'] = 'bar'
                ds1.to_netcdf(tmp1)
                ds2.to_netcdf(tmp2)
                with open_mfdataset([tmp1, tmp2]) as actual:
                    # presumes that attributes inherited from
                    # first dataset loaded
                    self.assertEqual(actual.test1, ds1.test1)
                    # attributes from ds2 are not retained, e.g.,
                    with raises_regex(AttributeError,
                                      'no attribute'):
                        actual.test2

    def test_preprocess_mfdataset(self):
        original = Dataset({'foo': ('x', np.random.randn(10))})
        with create_tmp_file() as tmp:
            original.to_netcdf(tmp)

            def preprocess(ds):
                return ds.assign_coords(z=0)

            expected = preprocess(original)
            with open_mfdataset(tmp, preprocess=preprocess,
                                autoclose=self.autoclose) as actual:
                assert_identical(expected, actual)

    def test_save_mfdataset_roundtrip(self):
        original = Dataset({'foo': ('x', np.random.randn(10))})
        datasets = [original.isel(x=slice(5)),
                    original.isel(x=slice(5, 10))]
        with create_tmp_file() as tmp1:
            with create_tmp_file() as tmp2:
                save_mfdataset(datasets, [tmp1, tmp2])
                with open_mfdataset([tmp1, tmp2],
                                    autoclose=self.autoclose) as actual:
                    assert_identical(actual, original)

    def test_save_mfdataset_invalid(self):
        ds = Dataset()
        with raises_regex(ValueError, 'cannot use mode'):
            save_mfdataset([ds, ds], ['same', 'same'])
        with raises_regex(ValueError, 'same length'):
            save_mfdataset([ds, ds], ['only one path'])

    def test_save_mfdataset_invalid_dataarray(self):
        # regression test for GH1555
        da = DataArray([1, 2])
        with raises_regex(TypeError, 'supports writing Dataset'):
            save_mfdataset([da], ['dataarray'])

    @requires_pathlib
    def test_save_mfdataset_pathlib_roundtrip(self):
        original = Dataset({'foo': ('x', np.random.randn(10))})
        datasets = [original.isel(x=slice(5)),
                    original.isel(x=slice(5, 10))]
        with create_tmp_file() as tmp1:
            with create_tmp_file() as tmp2:
                tmp1 = Path(tmp1)
                tmp2 = Path(tmp2)
                save_mfdataset(datasets, [tmp1, tmp2])
                with open_mfdataset([tmp1, tmp2],
                                    autoclose=self.autoclose) as actual:
                    assert_identical(actual, original)

    def test_open_and_do_math(self):
        original = Dataset({'foo': ('x', np.random.randn(10))})
        with create_tmp_file() as tmp:
            original.to_netcdf(tmp)
            with open_mfdataset(tmp, autoclose=self.autoclose) as ds:
                actual = 1.0 * ds
                assert_allclose(original, actual, decode_bytes=False)

    def test_open_mfdataset_concat_dim_none(self):
        with create_tmp_file() as tmp1:
            with create_tmp_file() as tmp2:
                data = Dataset({'x': 0})
                data.to_netcdf(tmp1)
                Dataset({'x': np.nan}).to_netcdf(tmp2)
                with open_mfdataset([tmp1, tmp2], concat_dim=None,
                                    autoclose=self.autoclose) as actual:
                    assert_identical(data, actual)

    def test_open_dataset(self):
        original = Dataset({'foo': ('x', np.random.randn(10))})
        with create_tmp_file() as tmp:
            original.to_netcdf(tmp)
            with open_dataset(tmp, chunks={'x': 5}) as actual:
                self.assertIsInstance(actual.foo.variable.data, da.Array)
                self.assertEqual(actual.foo.variable.data.chunks, ((5, 5),))
                assert_identical(original, actual)
            with open_dataset(tmp, chunks=5) as actual:
                assert_identical(original, actual)
            with open_dataset(tmp) as actual:
                self.assertIsInstance(actual.foo.variable.data, np.ndarray)
                assert_identical(original, actual)

    def test_open_single_dataset(self):
        # Test for issue GH #1988. This makes sure that the
        # concat_dim is utilized when specified in open_mfdataset().
        rnddata = np.random.randn(10)
        original = Dataset({'foo': ('x', rnddata)})
        dim = DataArray([100], name='baz', dims='baz')
        expected = Dataset({'foo': (('baz', 'x'), rnddata[np.newaxis, :])},
                           {'baz': [100]})
        with create_tmp_file() as tmp:
            original.to_netcdf(tmp)
            with open_mfdataset([tmp], concat_dim=dim,
                                autoclose=self.autoclose) as actual:
                assert_identical(expected, actual)

    def test_dask_roundtrip(self):
        with create_tmp_file() as tmp:
            data = create_test_data()
            data.to_netcdf(tmp)
            chunks = {'dim1': 4, 'dim2': 4, 'dim3': 4, 'time': 10}
            with open_dataset(tmp, chunks=chunks) as dask_ds:
                assert_identical(data, dask_ds)
                with create_tmp_file() as tmp2:
                    dask_ds.to_netcdf(tmp2)
                    with open_dataset(tmp2) as on_disk:
                        assert_identical(data, on_disk)

    def test_deterministic_names(self):
        with create_tmp_file() as tmp:
            data = create_test_data()
            data.to_netcdf(tmp)
            with open_mfdataset(tmp, autoclose=self.autoclose) as ds:
                original_names = dict((k, v.data.name)
                                      for k, v in ds.data_vars.items())
            with open_mfdataset(tmp, autoclose=self.autoclose) as ds:
                repeat_names = dict((k, v.data.name)
                                    for k, v in ds.data_vars.items())
            for var_name, dask_name in original_names.items():
                self.assertIn(var_name, dask_name)
                self.assertEqual(dask_name[:13], 'open_dataset-')
            self.assertEqual(original_names, repeat_names)

    def test_dataarray_compute(self):
        # Test DataArray.compute() on dask backend.
        # The test for Dataset.compute() is already in DatasetIOTestCases;
        # however dask is the only tested backend which supports DataArrays
        actual = DataArray([1, 2]).chunk()
        computed = actual.compute()
        self.assertFalse(actual._in_memory)
        self.assertTrue(computed._in_memory)
        assert_allclose(actual, computed, decode_bytes=False)

    def test_to_netcdf_compute_false_roundtrip(self):
        from dask.delayed import Delayed

        original = create_test_data().chunk()

        with create_tmp_file() as tmp_file:
            # dataset, path, **kwargs):
            delayed_obj = self.save(original, tmp_file, compute=False)
            assert isinstance(delayed_obj, Delayed)
            delayed_obj.compute()

            with self.open(tmp_file) as actual:
                assert_identical(original, actual)

    def test_save_mfdataset_compute_false_roundtrip(self):
        from dask.delayed import Delayed

        original = Dataset({'foo': ('x', np.random.randn(10))}).chunk()
        datasets = [original.isel(x=slice(5)),
                    original.isel(x=slice(5, 10))]
        with create_tmp_file() as tmp1:
            with create_tmp_file() as tmp2:
                delayed_obj = save_mfdataset(datasets, [tmp1, tmp2],
                                             engine=self.engine, compute=False)
                assert isinstance(delayed_obj, Delayed)
                delayed_obj.compute()
                with open_mfdataset([tmp1, tmp2],
                                    autoclose=self.autoclose) as actual:
                    assert_identical(actual, original)


class DaskTestAutocloseTrue(DaskTest):
    autoclose = True


@requires_scipy_or_netCDF4
@requires_pydap
class PydapTest(TestCase):
    def convert_to_pydap_dataset(self, original):
        from pydap.model import GridType, BaseType, DatasetType
        ds = DatasetType('bears', **original.attrs)
        for key, var in original.data_vars.items():
            v = GridType(key)
            v[key] = BaseType(key, var.values, dimensions=var.dims,
                              **var.attrs)
            for d in var.dims:
                v[d] = BaseType(d, var[d].values)
            ds[key] = v
        # check all dims are stored in ds
        for d in original.coords:
            ds[d] = BaseType(d, original[d].values, dimensions=(d, ),
                             **original[d].attrs)
        return ds

    @contextlib.contextmanager
    def create_datasets(self, **kwargs):
        with open_example_dataset('bears.nc') as expected:
            pydap_ds = self.convert_to_pydap_dataset(expected)
            actual = open_dataset(PydapDataStore(pydap_ds))
            # TODO solve this workaround:
            # netcdf converts string to byte not unicode
            expected['bears'] = expected['bears'].astype(str)
            yield actual, expected

    def test_cmp_local_file(self):
        with self.create_datasets() as (actual, expected):
            assert_equal(actual, expected)

            # global attributes should be global attributes on the dataset
            self.assertNotIn('NC_GLOBAL', actual.attrs)
            self.assertIn('history', actual.attrs)

            # we don't check attributes exactly with assertDatasetIdentical()
            # because the test DAP server seems to insert some extra
            # attributes not found in the netCDF file.
            assert actual.attrs.keys() == expected.attrs.keys()

        with self.create_datasets() as (actual, expected):
            assert_equal(
                actual.isel(l=2), expected.isel(l=2))  # noqa: E741

        with self.create_datasets() as (actual, expected):
            assert_equal(actual.isel(i=0, j=-1),
                         expected.isel(i=0, j=-1))

        with self.create_datasets() as (actual, expected):
            assert_equal(actual.isel(j=slice(1, 2)),
                         expected.isel(j=slice(1, 2)))

        with self.create_datasets() as (actual, expected):
            indexers = {'i': [1, 0, 0], 'j': [1, 2, 0, 1]}
            assert_equal(actual.isel(**indexers),
                         expected.isel(**indexers))

        with self.create_datasets() as (actual, expected):
            indexers = {'i': DataArray([0, 1, 0], dims='a'),
                        'j': DataArray([0, 2, 1], dims='a')}
            assert_equal(actual.isel(**indexers),
                         expected.isel(**indexers))

    def test_compatible_to_netcdf(self):
        # make sure it can be saved as a netcdf
        with self.create_datasets() as (actual, expected):
            with create_tmp_file() as tmp_file:
                actual.to_netcdf(tmp_file)
                actual = open_dataset(tmp_file)
                actual['bears'] = actual['bears'].astype(str)
                assert_equal(actual, expected)

    @requires_dask
    def test_dask(self):
        with self.create_datasets(chunks={'j': 2}) as (actual, expected):
            assert_equal(actual, expected)


@network
@requires_scipy_or_netCDF4
@requires_pydap
class PydapOnlineTest(PydapTest):
    @contextlib.contextmanager
    def create_datasets(self, **kwargs):
        url = 'http://test.opendap.org/opendap/hyrax/data/nc/bears.nc'
        actual = open_dataset(url, engine='pydap', **kwargs)
        with open_example_dataset('bears.nc') as expected:
            # workaround to restore string which is converted to byte
            expected['bears'] = expected['bears'].astype(str)
            yield actual, expected

    def test_session(self):
        from pydap.cas.urs import setup_session

        session = setup_session('XarrayTestUser', 'Xarray2017')
        with mock.patch('pydap.client.open_url') as mock_func:
            xr.backends.PydapDataStore.open('http://test.url', session=session)
        mock_func.assert_called_with('http://test.url', session=session)


@requires_scipy
@requires_pynio
class PyNioTest(ScipyWriteTest, TestCase):
    def test_write_store(self):
        # pynio is read-only for now
        pass

    @contextlib.contextmanager
    def open(self, path, **kwargs):
        with open_dataset(path, engine='pynio', autoclose=self.autoclose,
                          **kwargs) as ds:
            yield ds

    def save(self, dataset, path, **kwargs):
        return dataset.to_netcdf(path, engine='scipy', **kwargs)

    def test_weakrefs(self):
        example = Dataset({'foo': ('x', np.arange(5.0))})
        expected = example.rename({'foo': 'bar', 'x': 'y'})

        with create_tmp_file() as tmp_file:
            example.to_netcdf(tmp_file, engine='scipy')
            on_disk = open_dataset(tmp_file, engine='pynio')
            actual = on_disk.rename({'foo': 'bar', 'x': 'y'})
            del on_disk  # trigger garbage collection
            assert_identical(actual, expected)


class PyNioTestAutocloseTrue(PyNioTest):
    autoclose = True


@requires_pseudonetcdf
class PseudoNetCDFFormatTest(TestCase):
    autoclose = True

    def open(self, path, **kwargs):
        return open_dataset(path, engine='pseudonetcdf',
                            autoclose=self.autoclose,
                            **kwargs)

    @contextlib.contextmanager
    def roundtrip(self, data, save_kwargs={}, open_kwargs={},
                  allow_cleanup_failure=False):
        with create_tmp_file(
                allow_cleanup_failure=allow_cleanup_failure) as path:
            self.save(data, path, **save_kwargs)
            with self.open(path, **open_kwargs) as ds:
                yield ds

    def test_ict_format(self):
        """
        Open a CAMx file and test data variables
        """
        ictfile = open_example_dataset('example.ict',
                                       engine='pseudonetcdf',
                                       autoclose=False,
                                       backend_kwargs={'format': 'ffi1001'})
        stdattr = {
            'fill_value': -9999.0,
            'missing_value': -9999,
            'scale': 1,
            'llod_flag': -8888,
            'llod_value': 'N/A',
            'ulod_flag': -7777,
            'ulod_value': 'N/A'
        }

        def myatts(**attrs):
            outattr = stdattr.copy()
            outattr.update(attrs)
            return outattr

        input = {
            'coords': {},
            'attrs': {
                'fmt': '1001', 'n_header_lines': 27,
                'PI_NAME': 'Henderson, Barron',
                'ORGANIZATION_NAME': 'U.S. EPA',
                'SOURCE_DESCRIPTION': 'Example file with artificial data',
                'MISSION_NAME': 'JUST_A_TEST',
                'VOLUME_INFO': '1, 1',
                'SDATE': '2018, 04, 27', 'WDATE': '2018, 04, 27',
                'TIME_INTERVAL': '0',
                'INDEPENDENT_VARIABLE': 'Start_UTC',
                'ULOD_FLAG': '-7777', 'ULOD_VALUE': 'N/A',
                'LLOD_FLAG': '-8888',
                'LLOD_VALUE': ('N/A, N/A, N/A, N/A, 0.025'),
                'OTHER_COMMENTS': ('www-air.larc.nasa.gov/missions/etc/' +
                                   'IcarttDataFormat.htm'),
                'REVISION': 'R0',
                'R0': 'No comments for this revision.',
                'TFLAG': 'Start_UTC'
            },
            'dims': {'POINTS': 4},
            'data_vars': {
                'Start_UTC': {
                    'data': [43200.0, 46800.0, 50400.0, 50400.0],
                    'dims': ('POINTS',),
                    'attrs': myatts(
                        units='Start_UTC',
                        standard_name='Start_UTC',
                    )
                },
                'lat': {
                    'data': [41.0, 42.0, 42.0, 42.0],
                    'dims': ('POINTS',),
                    'attrs': myatts(
                        units='degrees_north',
                        standard_name='lat',
                    )
                },
                'lon': {
                    'data': [-71.0, -72.0, -73.0, -74.],
                    'dims': ('POINTS',),
                    'attrs': myatts(
                        units='degrees_east',
                        standard_name='lon',
                    )
                },
                'elev': {
                    'data': [5.0, 15.0, 20.0, 25.0],
                    'dims': ('POINTS',),
                    'attrs': myatts(
                        units='meters',
                        standard_name='elev',
                    )
                },
                'TEST_ppbv': {
                    'data': [1.2345, 2.3456, 3.4567, 4.5678],
                    'dims': ('POINTS',),
                    'attrs': myatts(
                        units='ppbv',
                        standard_name='TEST_ppbv',
                    )
                },
                'TESTM_ppbv': {
                    'data': [2.22, -9999.0, -7777.0, -8888.0],
                    'dims': ('POINTS',),
                    'attrs': myatts(
                        units='ppbv',
                        standard_name='TESTM_ppbv',
                        llod_value=0.025
                    )
                }
            }
        }
        chkfile = Dataset.from_dict(input)
        assert_identical(ictfile, chkfile)

    def test_ict_format_write(self):
        fmtkw = {'format': 'ffi1001'}
        expected = open_example_dataset('example.ict',
                                        engine='pseudonetcdf',
                                        autoclose=False,
                                        backend_kwargs=fmtkw)
        with self.roundtrip(expected, save_kwargs=fmtkw,
                            open_kwargs={'backend_kwargs': fmtkw}) as actual:
            assert_identical(expected, actual)

    def test_uamiv_format_read(self):
        """
        Open a CAMx file and test data variables
        """
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning,
                                    message=('IOAPI_ISPH is assumed to be ' +
                                             '6370000.; consistent with WRF'))
            camxfile = open_example_dataset('example.uamiv',
                                            engine='pseudonetcdf',
                                            autoclose=True,
                                            backend_kwargs={'format': 'uamiv'})
        data = np.arange(20, dtype='f').reshape(1, 1, 4, 5)
        expected = xr.Variable(('TSTEP', 'LAY', 'ROW', 'COL'), data,
                               dict(units='ppm', long_name='O3'.ljust(16),
                                    var_desc='O3'.ljust(80)))
        actual = camxfile.variables['O3']
        assert_allclose(expected, actual)

        data = np.array(['2002-06-03'], 'datetime64[ns]')
        expected = xr.Variable(('TSTEP',), data,
                               dict(bounds='time_bounds',
                                    long_name=('synthesized time coordinate ' +
                                               'from SDATE, STIME, STEP ' +
                                               'global attributes')))
        actual = camxfile.variables['time']
        assert_allclose(expected, actual)
        camxfile.close()

    def test_uamiv_format_mfread(self):
        """
        Open a CAMx file and test data variables
        """
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning,
                                    message=('IOAPI_ISPH is assumed to be ' +
                                             '6370000.; consistent with WRF'))
            camxfile = open_example_mfdataset(
                ['example.uamiv',
                 'example.uamiv'],
                engine='pseudonetcdf',
                autoclose=True,
                concat_dim='TSTEP',
                backend_kwargs={'format': 'uamiv'})

        data1 = np.arange(20, dtype='f').reshape(1, 1, 4, 5)
        data = np.concatenate([data1] * 2, axis=0)
        expected = xr.Variable(('TSTEP', 'LAY', 'ROW', 'COL'), data,
                               dict(units='ppm', long_name='O3'.ljust(16),
                                    var_desc='O3'.ljust(80)))
        actual = camxfile.variables['O3']
        assert_allclose(expected, actual)

        data1 = np.array(['2002-06-03'], 'datetime64[ns]')
        data = np.concatenate([data1] * 2, axis=0)
        expected = xr.Variable(('TSTEP',), data,
                               dict(bounds='time_bounds',
                                    long_name=('synthesized time coordinate ' +
                                               'from SDATE, STIME, STEP ' +
                                               'global attributes')))
        actual = camxfile.variables['time']
        assert_allclose(expected, actual)
        camxfile.close()

    def test_uamiv_format_write(self):
        fmtkw = {'format': 'uamiv'}
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning,
                                    message=('IOAPI_ISPH is assumed to be ' +
                                             '6370000.; consistent with WRF'))
            expected = open_example_dataset('example.uamiv',
                                            engine='pseudonetcdf',
                                            autoclose=False,
                                            backend_kwargs=fmtkw)
        with self.roundtrip(expected,
                            save_kwargs=fmtkw,
                            open_kwargs={'backend_kwargs': fmtkw}) as actual:
            assert_identical(expected, actual)

    def save(self, dataset, path, **save_kwargs):
        import PseudoNetCDF as pnc
        pncf = pnc.PseudoNetCDFFile()
        pncf.dimensions = {k: pnc.PseudoNetCDFDimension(pncf, k, v)
                           for k, v in dataset.dims.items()}
        pncf.variables = {k: pnc.PseudoNetCDFVariable(pncf, k, v.dtype.char,
                                                      v.dims,
                                                      values=v.data[...],
                                                      **v.attrs)
                          for k, v in dataset.variables.items()}
        for pk, pv in dataset.attrs.items():
            setattr(pncf, pk, pv)

        pnc.pncwrite(pncf, path, **save_kwargs)


@requires_rasterio
@contextlib.contextmanager
def create_tmp_geotiff(nx=4, ny=3, nz=3,
                       transform=None,
                       transform_args=[5000, 80000, 1000, 2000.],
                       crs={'units': 'm', 'no_defs': True, 'ellps': 'WGS84',
                            'proj': 'utm', 'zone': 18},
                       open_kwargs={}):
    # yields a temporary geotiff file and a corresponding expected DataArray
    import rasterio
    from rasterio.transform import from_origin
    with create_tmp_file(suffix='.tif',
                         allow_cleanup_failure=ON_WINDOWS) as tmp_file:
        # allow 2d or 3d shapes
        if nz == 1:
            data_shape = ny, nx
            write_kwargs = {'indexes': 1}
        else:
            data_shape = nz, ny, nx
            write_kwargs = {}
        data = np.arange(
            nz * ny * nx,
            dtype=rasterio.float32).reshape(
            *data_shape)
        if transform is None:
            transform = from_origin(*transform_args)
        with rasterio.open(
                tmp_file, 'w',
                driver='GTiff', height=ny, width=nx, count=nz,
                crs=crs,
                transform=transform,
                dtype=rasterio.float32,
                **open_kwargs) as s:
            s.write(data, **write_kwargs)
            dx, dy = s.res[0], -s.res[1]

        a, b, c, d = transform_args
        data = data[np.newaxis, ...] if nz == 1 else data
        expected = DataArray(data, dims=('band', 'y', 'x'),
                             coords={
                                 'band': np.arange(nz) + 1,
                                 'y': -np.arange(ny) * d + b + dy / 2,
                                 'x': np.arange(nx) * c + a + dx / 2,
        })
        yield tmp_file, expected


@requires_rasterio
class TestRasterio(TestCase):

    @requires_scipy_or_netCDF4
    def test_serialization(self):
        with create_tmp_geotiff() as (tmp_file, expected):
            # Write it to a netcdf and read again (roundtrip)
            with xr.open_rasterio(tmp_file) as rioda:
                with create_tmp_file(suffix='.nc') as tmp_nc_file:
                    rioda.to_netcdf(tmp_nc_file)
                    with xr.open_dataarray(tmp_nc_file) as ncds:
                        assert_identical(rioda, ncds)

    def test_utm(self):
        with create_tmp_geotiff() as (tmp_file, expected):
            with xr.open_rasterio(tmp_file) as rioda:
                assert_allclose(rioda, expected)
                assert isinstance(rioda.attrs['crs'], basestring)
                assert isinstance(rioda.attrs['res'], tuple)
                assert isinstance(rioda.attrs['is_tiled'], np.uint8)
                assert isinstance(rioda.attrs['transform'], tuple)
                np.testing.assert_array_equal(rioda.attrs['nodatavals'],
                                              [np.NaN, np.NaN, np.NaN])

            # Check no parse coords
            with xr.open_rasterio(tmp_file, parse_coordinates=False) as rioda:
                assert 'x' not in rioda.coords
                assert 'y' not in rioda.coords

    def test_non_rectilinear(self):
        from rasterio.transform import from_origin
        # Create a geotiff file with 2d coordinates
        with create_tmp_geotiff(transform=from_origin(0, 3, 1, 1).rotation(45),
                                crs=None) as (tmp_file, _):
            # Default is to not parse coords
            with xr.open_rasterio(tmp_file) as rioda:
                assert 'x' not in rioda.coords
                assert 'y' not in rioda.coords
                assert 'crs' not in rioda.attrs
                assert isinstance(rioda.attrs['res'], tuple)
                assert isinstance(rioda.attrs['is_tiled'], np.uint8)
                assert isinstance(rioda.attrs['transform'], tuple)

            # See if a warning is raised if we force it
            with self.assertWarns("transformation isn't rectilinear"):
                with xr.open_rasterio(tmp_file,
                                      parse_coordinates=True) as rioda:
                    assert 'x' not in rioda.coords
                    assert 'y' not in rioda.coords

    def test_platecarree(self):
        with create_tmp_geotiff(8, 10, 1, transform_args=[1, 2, 0.5, 2.],
                                crs='+proj=latlong',
                                open_kwargs={'nodata': -9765}) \
                as (tmp_file, expected):
            with xr.open_rasterio(tmp_file) as rioda:
                assert_allclose(rioda, expected)
                assert isinstance(rioda.attrs['crs'], basestring)
                assert isinstance(rioda.attrs['res'], tuple)
                assert isinstance(rioda.attrs['is_tiled'], np.uint8)
                assert isinstance(rioda.attrs['transform'], tuple)
                np.testing.assert_array_equal(rioda.attrs['nodatavals'],
                                              [-9765.])

    def test_notransform(self):
        # regression test for https://github.com/pydata/xarray/issues/1686
        import rasterio
        import warnings

        # Create a geotiff file
        with warnings.catch_warnings():
            # rasterio throws a NotGeoreferencedWarning here, which is
            # expected since we test rasterio's defaults in this case.
            warnings.filterwarnings('ignore', category=UserWarning,
                                    message='Dataset has no geotransform set')
            with create_tmp_file(suffix='.tif') as tmp_file:
                # data
                nx, ny, nz = 4, 3, 3
                data = np.arange(nx * ny * nz,
                                 dtype=rasterio.float32).reshape(nz, ny, nx)
                with rasterio.open(
                        tmp_file, 'w',
                        driver='GTiff', height=ny, width=nx, count=nz,
                        dtype=rasterio.float32) as s:
                    s.write(data)

                # Tests
                expected = DataArray(data,
                                     dims=('band', 'y', 'x'),
                                     coords={'band': [1, 2, 3],
                                             'y': [0.5, 1.5, 2.5],
                                             'x': [0.5, 1.5, 2.5, 3.5],
                                             })
                with xr.open_rasterio(tmp_file) as rioda:
                    assert_allclose(rioda, expected)
                    assert isinstance(rioda.attrs['res'], tuple)
                    assert isinstance(rioda.attrs['is_tiled'], np.uint8)
                    assert isinstance(rioda.attrs['transform'], tuple)

    def test_indexing(self):
        with create_tmp_geotiff(8, 10, 3, transform_args=[1, 2, 0.5, 2.],
                                crs='+proj=latlong') as (tmp_file, expected):
            with xr.open_rasterio(tmp_file, cache=False) as actual:

                # tests
                # assert_allclose checks all data + coordinates
                assert_allclose(actual, expected)
                assert not actual.variable._in_memory

                # Basic indexer
                ind = {'x': slice(2, 5), 'y': slice(5, 7)}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                ind = {'band': slice(1, 2), 'x': slice(2, 5), 'y': slice(5, 7)}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                ind = {'band': slice(1, 2), 'x': slice(2, 5), 'y': 0}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                # orthogonal indexer
                ind = {'band': np.array([2, 1, 0]),
                       'x': np.array([1, 0]), 'y': np.array([0, 2])}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                ind = {'band': np.array([2, 1, 0]),
                       'x': np.array([1, 0]), 'y': 0}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                # minus-stepped slice
                ind = {'band': np.array([2, 1, 0]),
                       'x': slice(-1, None, -1), 'y': 0}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                ind = {'band': np.array([2, 1, 0]),
                       'x': 1, 'y': slice(-1, 1, -2)}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                # None is selected
                ind = {'band': np.array([2, 1, 0]),
                       'x': 1, 'y': slice(2, 2, 1)}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                # vectorized indexer
                ind = {'band': DataArray([2, 1, 0], dims='a'),
                       'x': DataArray([1, 0, 0], dims='a'),
                       'y': np.array([0, 2])}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                ind = {
                    'band': DataArray([[2, 1, 0], [1, 0, 2]], dims=['a', 'b']),
                    'x': DataArray([[1, 0, 0], [0, 1, 0]], dims=['a', 'b']),
                    'y': 0}
                assert_allclose(expected.isel(**ind), actual.isel(**ind))
                assert not actual.variable._in_memory

                # Selecting lists of bands is fine
                ex = expected.isel(band=[1, 2])
                ac = actual.isel(band=[1, 2])
                assert_allclose(ac, ex)
                ex = expected.isel(band=[0, 2])
                ac = actual.isel(band=[0, 2])
                assert_allclose(ac, ex)

                # Integer indexing
                ex = expected.isel(band=1)
                ac = actual.isel(band=1)
                assert_allclose(ac, ex)

                ex = expected.isel(x=1, y=2)
                ac = actual.isel(x=1, y=2)
                assert_allclose(ac, ex)

                ex = expected.isel(band=0, x=1, y=2)
                ac = actual.isel(band=0, x=1, y=2)
                assert_allclose(ac, ex)

                # Mixed
                ex = actual.isel(x=slice(2), y=slice(2))
                ac = actual.isel(x=[0, 1], y=[0, 1])
                assert_allclose(ac, ex)

                ex = expected.isel(band=0, x=1, y=slice(5, 7))
                ac = actual.isel(band=0, x=1, y=slice(5, 7))
                assert_allclose(ac, ex)

                ex = expected.isel(band=0, x=slice(2, 5), y=2)
                ac = actual.isel(band=0, x=slice(2, 5), y=2)
                assert_allclose(ac, ex)

                # One-element lists
                ex = expected.isel(band=[0], x=slice(2, 5), y=[2])
                ac = actual.isel(band=[0], x=slice(2, 5), y=[2])
                assert_allclose(ac, ex)

    def test_caching(self):
        with create_tmp_geotiff(8, 10, 3, transform_args=[1, 2, 0.5, 2.],
                                crs='+proj=latlong') as (tmp_file, expected):
            # Cache is the default
            with xr.open_rasterio(tmp_file) as actual:

                # This should cache everything
                assert_allclose(actual, expected)

                # once cached, non-windowed indexing should become possible
                ac = actual.isel(x=[2, 4])
                ex = expected.isel(x=[2, 4])
                assert_allclose(ac, ex)

    @requires_dask
    def test_chunks(self):
        with create_tmp_geotiff(8, 10, 3, transform_args=[1, 2, 0.5, 2.],
                                crs='+proj=latlong') as (tmp_file, expected):
            # Chunk at open time
            with xr.open_rasterio(tmp_file, chunks=(1, 2, 2)) as actual:

                import dask.array as da
                self.assertIsInstance(actual.data, da.Array)
                assert 'open_rasterio' in actual.data.name

                # do some arithmetic
                ac = actual.mean()
                ex = expected.mean()
                assert_allclose(ac, ex)

                ac = actual.sel(band=1).mean(dim='x')
                ex = expected.sel(band=1).mean(dim='x')
                assert_allclose(ac, ex)

    def test_pickle_rasterio(self):
        # regression test for https://github.com/pydata/xarray/issues/2121
        with create_tmp_geotiff() as (tmp_file, expected):
            with xr.open_rasterio(tmp_file) as rioda:
                temp = pickle.dumps(rioda)
                with pickle.loads(temp) as actual:
                    assert_equal(actual, rioda)

    def test_ENVI_tags(self):
        rasterio = pytest.importorskip('rasterio', minversion='1.0a')
        from rasterio.transform import from_origin

        # Create an ENVI file with some tags in the ENVI namespace
        # this test uses a custom driver, so we can't use create_tmp_geotiff
        with create_tmp_file(suffix='.dat') as tmp_file:
            # data
            nx, ny, nz = 4, 3, 3
            data = np.arange(nx * ny * nz,
                             dtype=rasterio.float32).reshape(nz, ny, nx)
            transform = from_origin(5000, 80000, 1000, 2000.)
            with rasterio.open(
                    tmp_file, 'w',
                    driver='ENVI', height=ny, width=nx, count=nz,
                    crs={'units': 'm', 'no_defs': True, 'ellps': 'WGS84',
                         'proj': 'utm', 'zone': 18},
                    transform=transform,
                    dtype=rasterio.float32) as s:
                s.update_tags(
                    ns='ENVI',
                    description='{Tagged file}',
                    wavelength='{123.000000, 234.234000, 345.345678}',
                    fwhm='{1.000000, 0.234000, 0.000345}')
                s.write(data)
                dx, dy = s.res[0], -s.res[1]

            # Tests
            coords = {
                'band': [1, 2, 3],
                'y': -np.arange(ny) * 2000 + 80000 + dy / 2,
                'x': np.arange(nx) * 1000 + 5000 + dx / 2,
                'wavelength': ('band', np.array([123, 234.234, 345.345678])),
                'fwhm': ('band', np.array([1, 0.234, 0.000345])),
            }
            expected = DataArray(data, dims=('band', 'y', 'x'), coords=coords)

            with xr.open_rasterio(tmp_file) as rioda:
                assert_allclose(rioda, expected)
                assert isinstance(rioda.attrs['crs'], basestring)
                assert isinstance(rioda.attrs['res'], tuple)
                assert isinstance(rioda.attrs['is_tiled'], np.uint8)
                assert isinstance(rioda.attrs['transform'], tuple)
                # from ENVI tags
                assert isinstance(rioda.attrs['description'], basestring)
                assert isinstance(rioda.attrs['map_info'], basestring)
                assert isinstance(rioda.attrs['samples'], basestring)

    def test_no_mftime(self):
        # rasterio can accept "filename" urguments that are actually urls,
        # including paths to remote files.
        # In issue #1816, we found that these caused dask to break, because
        # the modification time was used to determine the dask token. This
        # tests ensure we can still chunk such files when reading with
        # rasterio.
        with create_tmp_geotiff(8, 10, 3, transform_args=[1, 2, 0.5, 2.],
                                crs='+proj=latlong') as (tmp_file, expected):
            with mock.patch('os.path.getmtime', side_effect=OSError):
                with xr.open_rasterio(tmp_file, chunks=(1, 2, 2)) as actual:
                    import dask.array as da
                    self.assertIsInstance(actual.data, da.Array)
                    assert_allclose(actual, expected)

    @network
    def test_http_url(self):
        # more examples urls here
        # http://download.osgeo.org/geotiff/samples/
        url = 'http://download.osgeo.org/geotiff/samples/made_up/ntf_nord.tif'
        with xr.open_rasterio(url) as actual:
            assert actual.shape == (1, 512, 512)
        # make sure chunking works
        with xr.open_rasterio(url, chunks=(1, 256, 256)) as actual:
            import dask.array as da
            self.assertIsInstance(actual.data, da.Array)


class TestEncodingInvalid(TestCase):

    def test_extract_nc4_variable_encoding(self):
        var = xr.Variable(('x',), [1, 2, 3], {}, {'foo': 'bar'})
        with raises_regex(ValueError, 'unexpected encoding'):
            _extract_nc4_variable_encoding(var, raise_on_invalid=True)

        var = xr.Variable(('x',), [1, 2, 3], {}, {'chunking': (2, 1)})
        encoding = _extract_nc4_variable_encoding(var)
        self.assertEqual({}, encoding)

        # regression test
        var = xr.Variable(('x',), [1, 2, 3], {}, {'shuffle': True})
        encoding = _extract_nc4_variable_encoding(var, raise_on_invalid=True)
        self.assertEqual({'shuffle': True}, encoding)

    def test_extract_h5nc_encoding(self):
        # not supported with h5netcdf (yet)
        var = xr.Variable(('x',), [1, 2, 3], {},
                          {'least_sigificant_digit': 2})
        with raises_regex(ValueError, 'unexpected encoding'):
            _extract_nc4_variable_encoding(var, raise_on_invalid=True)


class MiscObject:
    pass


@requires_netCDF4
class TestValidateAttrs(TestCase):
    def test_validating_attrs(self):
        def new_dataset():
            return Dataset({'data': ('y', np.arange(10.0))},
                           {'y': np.arange(10)})

        def new_dataset_and_dataset_attrs():
            ds = new_dataset()
            return ds, ds.attrs

        def new_dataset_and_data_attrs():
            ds = new_dataset()
            return ds, ds.data.attrs

        def new_dataset_and_coord_attrs():
            ds = new_dataset()
            return ds, ds.coords['y'].attrs

        for new_dataset_and_attrs in [new_dataset_and_dataset_attrs,
                                      new_dataset_and_data_attrs,
                                      new_dataset_and_coord_attrs]:
            ds, attrs = new_dataset_and_attrs()

            attrs[123] = 'test'
            with raises_regex(TypeError, 'Invalid name for attr'):
                ds.to_netcdf('test.nc')

            ds, attrs = new_dataset_and_attrs()
            attrs[MiscObject()] = 'test'
            with raises_regex(TypeError, 'Invalid name for attr'):
                ds.to_netcdf('test.nc')

            ds, attrs = new_dataset_and_attrs()
            attrs[''] = 'test'
            with raises_regex(ValueError, 'Invalid name for attr'):
                ds.to_netcdf('test.nc')

            # This one should work
            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = 'test'
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = {'a': 5}
            with raises_regex(TypeError, 'Invalid value for attr'):
                ds.to_netcdf('test.nc')

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = MiscObject()
            with raises_regex(TypeError, 'Invalid value for attr'):
                ds.to_netcdf('test.nc')

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = 5
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = 3.14
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = [1, 2, 3, 4]
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = (1.9, 2.5)
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = np.arange(5)
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = np.arange(12).reshape(3, 4)
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = 'This is a string'
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = ''
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)

            ds, attrs = new_dataset_and_attrs()
            attrs['test'] = np.arange(12).reshape(3, 4)
            with create_tmp_file() as tmp_file:
                ds.to_netcdf(tmp_file)


@requires_scipy_or_netCDF4
class TestDataArrayToNetCDF(TestCase):

    def test_dataarray_to_netcdf_no_name(self):
        original_da = DataArray(np.arange(12).reshape((3, 4)))

        with create_tmp_file() as tmp:
            original_da.to_netcdf(tmp)

            with open_dataarray(tmp) as loaded_da:
                assert_identical(original_da, loaded_da)

    def test_dataarray_to_netcdf_with_name(self):
        original_da = DataArray(np.arange(12).reshape((3, 4)),
                                name='test')

        with create_tmp_file() as tmp:
            original_da.to_netcdf(tmp)

            with open_dataarray(tmp) as loaded_da:
                assert_identical(original_da, loaded_da)

    def test_dataarray_to_netcdf_coord_name_clash(self):
        original_da = DataArray(np.arange(12).reshape((3, 4)),
                                dims=['x', 'y'],
                                name='x')

        with create_tmp_file() as tmp:
            original_da.to_netcdf(tmp)

            with open_dataarray(tmp) as loaded_da:
                assert_identical(original_da, loaded_da)

    def test_open_dataarray_options(self):
        data = DataArray(
            np.arange(5), coords={'y': ('x', range(5))}, dims=['x'])

        with create_tmp_file() as tmp:
            data.to_netcdf(tmp)

            expected = data.drop('y')
            with open_dataarray(tmp, drop_variables=['y']) as loaded:
                assert_identical(expected, loaded)

    def test_dataarray_to_netcdf_return_bytes(self):
        # regression test for GH1410
        data = xr.DataArray([1, 2, 3])
        output = data.to_netcdf()
        assert isinstance(output, bytes)

    @requires_pathlib
    def test_dataarray_to_netcdf_no_name_pathlib(self):
        original_da = DataArray(np.arange(12).reshape((3, 4)))

        with create_tmp_file() as tmp:
            tmp = Path(tmp)
            original_da.to_netcdf(tmp)

            with open_dataarray(tmp) as loaded_da:
                assert_identical(original_da, loaded_da)


def test_pickle_reconstructor():

    lines = ['foo bar spam eggs']

    with create_tmp_file(allow_cleanup_failure=ON_WINDOWS) as tmp:
        with open(tmp, 'w') as f:
            f.writelines(lines)

        obj = PickleByReconstructionWrapper(open, tmp)

        assert obj.value.readlines() == lines

        p_obj = pickle.dumps(obj)
        obj.value.close()  # for windows
        obj2 = pickle.loads(p_obj)

        assert obj2.value.readlines() == lines

        # roundtrip again to make sure we can fully restore the state
        p_obj2 = pickle.dumps(obj2)
        obj2.value.close()  # for windows
        obj3 = pickle.loads(p_obj2)

        assert obj3.value.readlines() == lines


@requires_scipy_or_netCDF4
def test_no_warning_from_dask_effective_get():
    with create_tmp_file() as tmpfile:
        with pytest.warns(None) as record:
            ds = Dataset()
            ds.to_netcdf(tmpfile)
        assert len(record) == 0



from __future__ import absolute_import, division, print_function

from collections import Iterable
from functools import partial

import numpy as np
import pandas as pd

from . import rolling
from .computation import apply_ufunc
from .pycompat import iteritems
from .utils import is_scalar, OrderedSet
from .variable import Variable, broadcast_variables
from .duck_array_ops import dask_array_type


class BaseInterpolator(object):
    '''gerneric interpolator class for normalizing interpolation methods'''
    cons_kwargs = {}
    call_kwargs = {}
    f = None
    method = None

    def __init__(self, xi, yi, method=None, **kwargs):
        self.method = method
        self.call_kwargs = kwargs

    def __call__(self, x):
        return self.f(x, **self.call_kwargs)

    def __repr__(self):
        return "{type}: method={method}".format(type=self.__class__.__name__,
                                                method=self.method)


class NumpyInterpolator(BaseInterpolator):
    '''One-dimensional linear interpolation.

    See Also
    --------
    numpy.interp
    '''

    def __init__(self, xi, yi, method='linear', fill_value=None, **kwargs):

        if method != 'linear':
            raise ValueError(
                'only method `linear` is valid for the NumpyInterpolator')

        self.method = method
        self.f = np.interp
        self.cons_kwargs = kwargs
        self.call_kwargs = {'period': self.cons_kwargs.pop('period', None)}

        self._xi = xi
        self._yi = yi

        if self.cons_kwargs:
            raise ValueError(
                'recieved invalid kwargs: %r' % self.cons_kwargs.keys())

        if fill_value is None:
            self._left = np.nan
            self._right = np.nan
        elif isinstance(fill_value, Iterable) and len(fill_value) == 2:
            self._left = fill_value[0]
            self._right = fill_value[1]
        elif is_scalar(fill_value):
            self._left = fill_value
            self._right = fill_value
        else:
            raise ValueError('%s is not a valid fill_value' % fill_value)

    def __call__(self, x):
        return self.f(x, self._xi, self._yi, left=self._left,
                      right=self._right, **self.call_kwargs)


class ScipyInterpolator(BaseInterpolator):
    '''Interpolate a 1-D function using Scipy interp1d

    See Also
    --------
    scipy.interpolate.interp1d
    '''

    def __init__(self, xi, yi, method=None, fill_value=None,
                 assume_sorted=True, copy=False, bounds_error=False, **kwargs):
        from scipy.interpolate import interp1d

        if method is None:
            raise ValueError('method is a required argument, please supply a '
                             'valid scipy.inter1d method (kind)')

        if method == 'polynomial':
            method = kwargs.pop('order', None)
            if method is None:
                raise ValueError('order is required when method=polynomial')

        self.method = method

        self.cons_kwargs = kwargs
        self.call_kwargs = {}

        if fill_value is None and method == 'linear':
            fill_value = kwargs.pop('fill_value', (np.nan, np.nan))
        elif fill_value is None:
            fill_value = np.nan

        self.f = interp1d(xi, yi, kind=self.method, fill_value=fill_value,
                          bounds_error=False, assume_sorted=assume_sorted,
                          copy=copy, **self.cons_kwargs)


class SplineInterpolator(BaseInterpolator):
    '''One-dimensional smoothing spline fit to a given set of data points.

    See Also
    --------
    scipy.interpolate.UnivariateSpline
    '''

    def __init__(self, xi, yi, method='spline', fill_value=None, order=3,
                 **kwargs):
        from scipy.interpolate import UnivariateSpline

        if method != 'spline':
            raise ValueError(
                'only method `spline` is valid for the SplineInterpolator')

        self.method = method
        self.cons_kwargs = kwargs
        self.call_kwargs['nu'] = kwargs.pop('nu', 0)
        self.call_kwargs['ext'] = kwargs.pop('ext', None)

        if fill_value is not None:
            raise ValueError('SplineInterpolator does not support fill_value')

        self.f = UnivariateSpline(xi, yi, k=order, **self.cons_kwargs)


def _apply_over_vars_with_dim(func, self, dim=None, **kwargs):
    '''wrapper for datasets'''

    ds = type(self)(coords=self.coords, attrs=self.attrs)

    for name, var in iteritems(self.data_vars):
        if dim in var.dims:
            ds[name] = func(var, dim=dim, **kwargs)
        else:
            ds[name] = var

    return ds


def get_clean_interp_index(arr, dim, use_coordinate=True, **kwargs):
    '''get index to use for x values in interpolation.

    If use_coordinate is True, the coordinate that shares the name of the
    dimension along which interpolation is being performed will be used as the
    x values.

    If use_coordinate is False, the x values are set as an equally spaced
    sequence.
    '''
    if use_coordinate:
        if use_coordinate is True:
            index = arr.get_index(dim)
        else:
            index = arr.coords[use_coordinate]
            if index.ndim != 1:
                raise ValueError(
                    'Coordinates used for interpolation must be 1D, '
                    '%s is %dD.' % (use_coordinate, index.ndim))

        # raise if index cannot be cast to a float (e.g. MultiIndex)
        try:
            index = index.values.astype(np.float64)
        except (TypeError, ValueError):
            # pandas raises a TypeError
            # xarray/nuppy raise a ValueError
            raise TypeError('Index must be castable to float64 to support'
                            'interpolation, got: %s' % type(index))
        # check index sorting now so we can skip it later
        if not (np.diff(index) > 0).all():
            raise ValueError("Index must be monotonicly increasing")
    else:
        axis = arr.get_axis_num(dim)
        index = np.arange(arr.shape[axis], dtype=np.float64)

    return index


def interp_na(self, dim=None, use_coordinate=True, method='linear', limit=None,
              **kwargs):
    '''Interpolate values according to different methods.'''

    if dim is None:
        raise NotImplementedError('dim is a required argument')

    if limit is not None:
        valids = _get_valid_fill_mask(self, dim, limit)

    # method
    index = get_clean_interp_index(self, dim, use_coordinate=use_coordinate,
                                   **kwargs)
    interp_class, kwargs = _get_interpolator(method, **kwargs)
    interpolator = partial(func_interpolate_na, interp_class, **kwargs)

    arr = apply_ufunc(interpolator, index, self,
                      input_core_dims=[[dim], [dim]],
                      output_core_dims=[[dim]],
                      output_dtypes=[self.dtype],
                      dask='parallelized',
                      vectorize=True,
                      keep_attrs=True).transpose(*self.dims)

    if limit is not None:
        arr = arr.where(valids)

    return arr


def func_interpolate_na(interpolator, x, y, **kwargs):
    '''helper function to apply interpolation along 1 dimension'''
    # it would be nice if this wasn't necessary, works around:
    # "ValueError: assignment destination is read-only" in assignment below
    out = y.copy()

    nans = pd.isnull(y)
    nonans = ~nans

    # fast track for no-nans and all-nans cases
    n_nans = nans.sum()
    if n_nans == 0 or n_nans == len(y):
        return y

    f = interpolator(x[nonans], y[nonans], **kwargs)
    out[nans] = f(x[nans])
    return out


def _bfill(arr, n=None, axis=-1):
    '''inverse of ffill'''
    import bottleneck as bn

    arr = np.flip(arr, axis=axis)

    # fill
    arr = bn.push(arr, axis=axis, n=n)

    # reverse back to original
    return np.flip(arr, axis=axis)


def ffill(arr, dim=None, limit=None):
    '''forward fill missing values'''
    import bottleneck as bn

    axis = arr.get_axis_num(dim)

    # work around for bottleneck 178
    _limit = limit if limit is not None else arr.shape[axis]

    return apply_ufunc(bn.push, arr,
                       dask='parallelized',
                       keep_attrs=True,
                       output_dtypes=[arr.dtype],
                       kwargs=dict(n=_limit, axis=axis)).transpose(*arr.dims)


def bfill(arr, dim=None, limit=None):
    '''backfill missing values'''
    axis = arr.get_axis_num(dim)

    # work around for bottleneck 178
    _limit = limit if limit is not None else arr.shape[axis]

    return apply_ufunc(_bfill, arr,
                       dask='parallelized',
                       keep_attrs=True,
                       output_dtypes=[arr.dtype],
                       kwargs=dict(n=_limit, axis=axis)).transpose(*arr.dims)


def _get_interpolator(method, vectorizeable_only=False, **kwargs):
    '''helper function to select the appropriate interpolator class

    returns interpolator class and keyword arguments for the class
    '''
    interp1d_methods = ['linear', 'nearest', 'zero', 'slinear', 'quadratic',
                        'cubic', 'polynomial']
    valid_methods = interp1d_methods + ['barycentric', 'krog', 'pchip',
                                        'spline', 'akima']

    has_scipy = True
    try:
        from scipy import interpolate
    except ImportError:
        has_scipy = False

    # prioritize scipy.interpolate
    if (method == 'linear' and not
            kwargs.get('fill_value', None) == 'extrapolate' and
            not vectorizeable_only):
        kwargs.update(method=method)
        interp_class = NumpyInterpolator

    elif method in valid_methods:
        if not has_scipy:
            raise ImportError(
                'Interpolation with method `%s` requires scipy' % method)

        if method in interp1d_methods:
            kwargs.update(method=method)
            interp_class = ScipyInterpolator
        elif vectorizeable_only:
            raise ValueError('{} is not a vectorizeable interpolator. '
                             'Available methods are {}'.format(
                                 method, interp1d_methods))
        elif method == 'barycentric':
            interp_class = interpolate.BarycentricInterpolator
        elif method == 'krog':
            interp_class = interpolate.KroghInterpolator
        elif method == 'pchip':
            interp_class = interpolate.PchipInterpolator
        elif method == 'spline':
            kwargs.update(method=method)
            interp_class = SplineInterpolator
        elif method == 'akima':
            interp_class = interpolate.Akima1DInterpolator
        else:
            raise ValueError('%s is not a valid scipy interpolator' % method)
    else:
        raise ValueError('%s is not a valid interpolator' % method)

    return interp_class, kwargs


def _get_interpolator_nd(method, **kwargs):
    '''helper function to select the appropriate interpolator class

    returns interpolator class and keyword arguments for the class
    '''
    valid_methods = ['linear', 'nearest']

    try:
        from scipy import interpolate
    except ImportError:
        raise ImportError(
            'Interpolation with method `%s` requires scipy' % method)

    if method in valid_methods:
        kwargs.update(method=method)
        interp_class = interpolate.interpn
    else:
        raise ValueError('%s is not a valid interpolator for interpolating '
                         'over multiple dimensions.' % method)

    return interp_class, kwargs


def _get_valid_fill_mask(arr, dim, limit):
    '''helper function to determine values that can be filled when limit is not
    None'''
    kw = {dim: limit + 1}
    # we explicitly use construct method to avoid copy.
    new_dim = rolling._get_new_dimname(arr.dims, '_window')
    return (arr.isnull().rolling(min_periods=1, **kw)
            .construct(new_dim, fill_value=False)
            .sum(new_dim, skipna=False)) <= limit


def _assert_single_chunk(var, axes):
    for axis in axes:
        if len(var.chunks[axis]) > 1 or var.chunks[axis][0] < var.shape[axis]:
            raise NotImplementedError(
                'Chunking along the dimension to be interpolated '
                '({}) is not yet supported.'.format(axis))


def _localize(var, indexes_coords):
    """ Speed up for linear and nearest neighbor method.
    Only consider a subspace that is needed for the interpolation
    """
    indexes = {}
    for dim, [x, new_x] in indexes_coords.items():
        index = x.to_index()
        imin = index.get_loc(np.min(new_x.values), method='nearest')
        imax = index.get_loc(np.max(new_x.values), method='nearest')

        indexes[dim] = slice(max(imin - 2, 0), imax + 2)
        indexes_coords[dim] = (x[indexes[dim]], new_x)
    return var.isel(**indexes), indexes_coords


def _floatize_x(x, new_x):
    """ Make x and new_x float.
    This is particulary useful for datetime dtype.
    x, new_x: tuple of np.ndarray
    """
    x = list(x)
    new_x = list(new_x)
    for i in range(len(x)):
        if x[i].dtype.kind in 'Mm':
            # Scipy casts coordinates to np.float64, which is not accurate
            # enough for datetime64 (uses 64bit integer).
            # We assume that the most of the bits are used to represent the
            # offset (min(x)) and the variation (x - min(x)) can be
            # represented by float.
            xmin = np.min(x[i])
            x[i] = (x[i] - xmin).astype(np.float64)
            new_x[i] = (new_x[i] - xmin).astype(np.float64)
    return x, new_x


def interp(var, indexes_coords, method, **kwargs):
    """ Make an interpolation of Variable

    Parameters
    ----------
    var: Variable
    index_coords:
        Mapping from dimension name to a pair of original and new coordinates.
        Original coordinates should be sorted in strictly ascending order.
        Note that all the coordinates should be Variable objects.
    method: string
        One of {'linear', 'nearest', 'zero', 'slinear', 'quadratic',
        'cubic'}. For multidimensional interpolation, only
        {'linear', 'nearest'} can be used.
    **kwargs:
        keyword arguments to be passed to scipy.interpolate

    Returns
    -------
    Interpolated Variable

    See Also
    --------
    DataArray.interp
    Dataset.interp
    """
    if not indexes_coords:
        return var.copy()

    # simple speed up for the local interpolation
    if method in ['linear', 'nearest']:
        var, indexes_coords = _localize(var, indexes_coords)

    # default behavior
    kwargs['bounds_error'] = kwargs.get('bounds_error', False)

    # target dimensions
    dims = list(indexes_coords)
    x, new_x = zip(*[indexes_coords[d] for d in dims])
    destination = broadcast_variables(*new_x)

    # transpose to make the interpolated axis to the last position
    broadcast_dims = [d for d in var.dims if d not in dims]
    original_dims = broadcast_dims + dims
    new_dims = broadcast_dims + list(destination[0].dims)
    interped = interp_func(var.transpose(*original_dims).data,
                           x, destination, method, kwargs)

    result = Variable(new_dims, interped, attrs=var.attrs)

    # dimension of the output array
    out_dims = OrderedSet()
    for d in var.dims:
        if d in dims:
            out_dims.update(indexes_coords[d][1].dims)
        else:
            out_dims.add(d)
    return result.transpose(*tuple(out_dims))


def interp_func(var, x, new_x, method, kwargs):
    """
    multi-dimensional interpolation for array-like. Interpolated axes should be
    located in the last position.

    Parameters
    ----------
    var: np.ndarray or dask.array.Array
        Array to be interpolated. The final dimension is interpolated.
    x: a list of 1d array.
        Original coordinates. Should not contain NaN.
    new_x: a list of 1d array
        New coordinates. Should not contain NaN.
    method: string
        {'linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic'} for
        1-dimensional itnterpolation.
        {'linear', 'nearest'} for multidimensional interpolation
    **kwargs:
        Optional keyword arguments to be passed to scipy.interpolator

    Returns
    -------
    interpolated: array
        Interpolated array

    Note
    ----
    This requiers scipy installed.

    See Also
    --------
    scipy.interpolate.interp1d
    """
    if not x:
        return var.copy()

    if len(x) == 1:
        func, kwargs = _get_interpolator(method, vectorizeable_only=True,
                                         **kwargs)
    else:
        func, kwargs = _get_interpolator_nd(method, **kwargs)

    if isinstance(var, dask_array_type):
        import dask.array as da

        _assert_single_chunk(var, range(var.ndim - len(x), var.ndim))
        chunks = var.chunks[:-len(x)] + new_x[0].shape
        drop_axis = range(var.ndim - len(x), var.ndim)
        new_axis = range(var.ndim - len(x), var.ndim - len(x) + new_x[0].ndim)
        return da.map_blocks(_interpnd, var, x, new_x, func, kwargs,
                             dtype=var.dtype, chunks=chunks,
                             new_axis=new_axis, drop_axis=drop_axis)

    return _interpnd(var, x, new_x, func, kwargs)


def _interp1d(var, x, new_x, func, kwargs):
    # x, new_x are tuples of size 1.
    x, new_x = x[0], new_x[0]
    rslt = func(x, var, assume_sorted=True, **kwargs)(np.ravel(new_x))
    if new_x.ndim > 1:
        return rslt.reshape(var.shape[:-1] + new_x.shape)
    if new_x.ndim == 0:
        return rslt[..., -1]
    return rslt


def _interpnd(var, x, new_x, func, kwargs):
    x, new_x = _floatize_x(x, new_x)

    if len(x) == 1:
        return _interp1d(var, x, new_x, func, kwargs)

    # move the interpolation axes to the start position
    var = var.transpose(range(-len(x), var.ndim - len(x)))
    # stack new_x to 1 vector, with reshape
    xi = np.stack([x1.values.ravel() for x1 in new_x], axis=-1)
    rslt = func(x, var, xi, **kwargs)
    # move back the interpolation axes to the last position
    rslt = rslt.transpose(range(-rslt.ndim + 1, 1))
    return rslt.reshape(rslt.shape[:-1] + new_x[0].shape)
